from datetime import date

from django.contrib.contenttypes.models import ContentType

from contract.tests.helpers import create_test_contract
from policyholder.tests.helpers import create_test_policy_holder
from insuree.test_helpers import create_test_insuree
from core.forms import User

from invoice.models import Invoice
from invoice.tests.helpers.default_test_data import DEFAULT_TEST_INVOICE_PAYLOAD


def __get_or_create_user():
    user = User.objects.filter(username='admin_invoice').first()
    if not user:
        user = User.objects.create_superuser(username='admin_invoice', password='S\/pe®Pąßw0rd™')
    return user


def create_test_invoice(subject=None, thirdparty=None, user=None, **custom_props):
    subject = subject or __create_test_subject()
    thirdparty = thirdparty or __create_test_thirdparty()
    payload = DEFAULT_TEST_INVOICE_PAYLOAD.copy()
    payload['subject'] = subject
    payload['subject_type'] = ContentType.objects.get_for_model(subject)
    payload['thirdparty'] = thirdparty
    payload['thirdparty_type'] = ContentType.objects.get_for_model(thirdparty)
    payload.update(**custom_props)
    user = user or __get_or_create_user()

    if Invoice.objects.filter(code=payload['code']).exists():
        i = Invoice.objects.filter(code=payload['code']).first()
        i.payments.all().delete()
        i.line_items.all().delete()
        Invoice.objects.filter(code=payload['code']).delete()

    invoice = Invoice(**payload)
    invoice.save(username=user.username)

    return invoice


def __create_test_subject():
    policy_holder = create_test_policy_holder()
    return create_test_contract(policy_holder)


def __create_test_thirdparty():
    return create_test_insuree(with_family=False)



from abc import ABC
from typing import Union, Iterable

from django.core.exceptions import ValidationError

from invoice.models import InvoicePayment, Invoice
from django.utils.translation import gettext as _


class GenericPaymentStatusValidation(ABC):
    @property
    def allowed_invoice_statuses(self) -> Iterable[Invoice.Status]:
        """
        Iterable of valid invoice statuses.
        """
        raise NotImplementedError()

    @property
    def allowed_payment_statuses(self):
        """
        Iterable of valid invoice payment statuses.
        """
        raise NotImplementedError()

    @property
    def error_message_invalid_invoice(self):
        """
        gettext string for content of ValidationError message raised when invoice status is invalid.
        """
        raise NotImplementedError()

    @property
    def error_message_invalid_payment(self):
        """
        gettext string for content of ValidationError message raised when invoice payment status is invalid.
        """
        raise NotImplementedError()

    @property
    def error_message_vars(self):
        """
        Returns dictionary of variables used by error messages.
        """
        invoice_status = self.invoice_payment.invoice.status
        payment_status = self.invoice_payment.status
        return {
            'invoice': self.invoice_payment.invoice,
            'payment': self.invoice_payment,
            'invoice_status':
                invoice_status.label if isinstance(invoice_status, Invoice.Status) else invoice_status,
            'payment_status':
                payment_status.label if isinstance(payment_status, InvoicePayment.PaymentStatus)
                else payment_status,
            'allowed_invoice': ', '.join([x.label for x in self.allowed_invoice_statuses]),
            'allowed_payment': ', '.join([x.label for x in self.allowed_payment_statuses]),
        }

    def __init__(self, invoice_payment):
        self.invoice_payment = invoice_payment

    def __call__(self, *args, **kwargs):
        self.validate()

    def validate(self):
        self._validate_invoice()
        self._validate_payment()

    @classmethod
    def _validate_status(
            cls, status: Union[InvoicePayment.PaymentStatus, Invoice.Status],
            allowed: Union[Iterable[InvoicePayment.PaymentStatus], Iterable[Invoice.Status]],
            err_msg):
        if status not in allowed:
            raise ValidationError(err_msg)

    def _validate_invoice(self):
        status = self.invoice_payment.invoice.status
        allowed = self.allowed_invoice_statuses
        err_msg = self.error_message_invalid_invoice % self.error_message_vars
        return self._validate_status(status, allowed, err_msg)

    def _validate_payment(self):
        status = self.invoice_payment.status
        allowed = self.allowed_payment_statuses
        err_msg = self.error_message_invalid_payment % self.error_message_vars
        return self._validate_status(status, allowed, err_msg)


class InvoicePaymentReceiveStatusValidator(GenericPaymentStatusValidation):
    allowed_invoice_statuses = [Invoice.Status.DRAFT, Invoice.Status.VALIDATED]

    error_message_invalid_invoice = _(
        "Payment for invoice %(invoice)s can't be made. Invoice has to be in status"
        "[%(allowed_invoice)s]. Invoice status is %(invoice_status)s."
    )

    allowed_payment_statuses = \
        [InvoicePayment.PaymentStatus.ACCEPTED, InvoicePayment.PaymentStatus.REJECTED]

    error_message_invalid_payment = _(
        "It's not possible to set payment status to %(payment_status)s during payment. Allowed statuses are "
        "%(allowed_payment)s."
    )


class InvoicePaymentRefundStatusValidator(GenericPaymentStatusValidation):
    allowed_invoice_statuses = [Invoice.Status.PAID]
    error_message_invalid_invoice = _(
        "Invoice %(invoice)s can't be refunded. Invoice has to be in PAYED status, currently it's %(invoice_status)s."
    )

    allowed_payment_statuses = [InvoicePayment.PaymentStatus.ACCEPTED]
    error_message_invalid_payment = _(
        "It's not possible to refund payment %(payment)s, as it's in %(payment_status) status. "
        "Only ACCEPTED payments can be refunded."
    )


class InvoicePaymentCancelStatusValidator(GenericPaymentStatusValidation):
    allowed_invoice_statuses = [Invoice.Status.PAID]

    error_message_invalid_invoice = _(
        "Invoice %(invoice)s can't be canceled. Invoice has to be in PAYED status, currently it's %(invoice_status)s."
    )

    allowed_payment_statuses = [InvoicePayment.PaymentStatus.ACCEPTED]
    error_message_invalid_payment = _(
        "It's not possible to cancel payment %(payment)s, only ACCEPTED payments can be canceled."
    )



from datetime import date
from enum import Enum
from typing import Union

from core.models import HistoryBusinessModel, HistoryModel, BaseVersionedModel
from invoice.models import Bill
from invoice.validation import BaseInvoiceValidation


class BillItemStatus(Enum):
    VALID = 0
    NO_ITEM = 1
    INVALID_ITEM = 2


class BillModelValidation(BaseInvoiceValidation):
    OBJECT_TYPE = Bill

    @classmethod
    def validate_subject(cls, bill: Bill):
        subject = bill.subject
        item_status = cls._get_item_status(subject)
        return item_status

    @classmethod
    def validate_line_items(cls, bill: Bill):
        statuses = {}
        for item in bill.line_items_bill.all():
            status = cls._get_item_status(item.line)
            statuses[item] = status
        return statuses

    @classmethod
    def _get_item_status(cls, subject: Union[HistoryModel, BaseVersionedModel]):
        if subject is None:
            return BillItemStatus.NO_ITEM
        else:
            if not cls.__valid_model(subject):
                return BillItemStatus.INVALID_ITEM
        return BillItemStatus.VALID

    @classmethod
    def __valid_model(cls, subject):
        # TODO: We should include generic method for checking if model is valid in both history model and
        # BaseVersionedModel
        if isinstance(subject, HistoryBusinessModel):
            return cls._validate_history_business_model(subject)
        elif isinstance(subject, HistoryModel):
            return cls._validate_history_model(subject)
        elif isinstance(subject, BaseVersionedModel):
            return cls._invoice_validate_base_versioned_model(subject)

    @classmethod
    def _validate_history_model(cls, subject: HistoryModel) -> bool:
        return not subject.is_deleted

    @classmethod
    def _validate_history_business_model(cls, subject: HistoryBusinessModel) -> bool:
        return cls._validate_history_model(subject) \
               and (subject.date_valid_to is None or subject.date_valid_to >= date.today())

    @classmethod
    def _invoice_validate_base_versioned_model(cls, subject: BaseVersionedModel) -> bool:
        return subject.validity_to is None



import graphene
from django.contrib.auth.models import AnonymousUser
from django.db.models import Q

from core.schema import OrderedDjangoFilterConnectionField
from core.utils import append_validity_filter
from invoice.apps import InvoiceConfig
from invoice.gql.gql_types.invoice_types import InvoiceGQLType
from invoice.models import Invoice
import graphene_django_optimizer as gql_optimizer


class InvoiceQueryMixin:
    invoice = OrderedDjangoFilterConnectionField(
        InvoiceGQLType,
        orderBy=graphene.List(of_type=graphene.String),
        dateValidFrom__Gte=graphene.DateTime(),
        dateValidTo__Lte=graphene.DateTime(),
        applyDefaultValidityFilter=graphene.Boolean(),
        client_mutation_id=graphene.String()
    )

    def resolve_invoice(self, info, **kwargs):
        InvoiceQueryMixin._check_permissions(info.context.user)
        filters = []
        filters += append_validity_filter(**kwargs)

        client_mutation_id = kwargs.get("client_mutation_id", None)
        if client_mutation_id:
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        subject_type = kwargs.get("subject_type", None)
        if subject_type:
            filters.append(Q(subject_type__model=subject_type))

        thirdparty_type = kwargs.get("thirdparty_type", None)
        if thirdparty_type:
            filters.append(Q(thirdparty_type__model=thirdparty_type))

        qs = Invoice.objects.filter(*filters)
        if InvoiceConfig.invoice_user_filter:
            qs = InvoiceConfig.invoice_user_filter(qs, info.context.user)

        return gql_optimizer.query(qs, info)

    @staticmethod
    def _check_permissions(user):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_invoice_search_perms):
            raise PermissionError("Unauthorized")



import graphene
import json
from django.core.serializers.json import DjangoJSONEncoder
from graphene_django import DjangoObjectType

from core import prefix_filterset, ExtendedConnection
from invoice.gql.filter_mixin import GenericFilterGQLTypeMixin
from invoice.models import Bill, \
    BillItem, BillEvent, BillPayment
from invoice.utils import underscore_to_camel
from location.models import Location


class BillGQLType(DjangoObjectType, GenericFilterGQLTypeMixin):
    subject = graphene.JSONString()
    subject_type = graphene.Int()
    subject_type_name = graphene.String()
    thirdparty = graphene.JSONString()
    thirdparty_type = graphene.Int()
    thirdparty_type_name = graphene.String()

    def resolve_subject_type(root, info):
        if root.subject_type:
            return root.subject_type.id

    def resolve_subject_type_name(root, info):
        if root.subject_type:
            return root.subject_type.name

    def resolve_thirdparty_type(root, info):
        if root.thirdparty_type:
            return root.thirdparty_type.id

    def resolve_thirdparty_type_name(root, info):
        if root.thirdparty_type:
            return root.thirdparty_type.name

    def resolve_subject(root, info):
        if root.subject:
            subject_object_dict = root.subject.__dict__
            subject_object_dict.pop('_state', None)
            subject_object_dict = {
                underscore_to_camel(k): v for k, v in list(subject_object_dict.items())
            }
            if root.subject_type.name == "batch run" and subject_object_dict.get('locationId', None):
                location = Location.objects.filter(id=subject_object_dict['locationId'], validity_to__isnull=True)
                location = location.values('code', 'name')
                subject_object_dict['location'] = {
                    underscore_to_camel(k): v for k, v in location.first().items()
                }
            subject_object_dict = json.dumps(subject_object_dict, cls=DjangoJSONEncoder)
            return subject_object_dict

    def resolve_thirdparty(root, info):
        if root.thirdparty:
            thirdparty_object_dict = root.thirdparty.__dict__
            thirdparty_object_dict.pop('_state', None)
            thirdparty_object_dict = {
                underscore_to_camel(k): v for k, v in list(thirdparty_object_dict.items())
            }
            thirdparty_object_dict = json.dumps(thirdparty_object_dict, cls=DjangoJSONEncoder)
            return thirdparty_object_dict

    class Meta:
        model = Bill
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            **GenericFilterGQLTypeMixin.get_base_filters_invoice(),
            "date_bill": ["exact", "lt", "lte", "gt", "gte"],
        }

        connection_class = ExtendedConnection

        @classmethod
        def get_queryset(cls, queryset, info):
            return Bill.get_queryset(queryset, info)


class BillItemGQLType(DjangoObjectType, GenericFilterGQLTypeMixin):
    line_type = graphene.Int()
    line_type_name = graphene.String()
    line = graphene.JSONString()

    def resolve_line_type(root, info):
        if root.line_type:
            return root.line_type.id

    def resolve_line_type_name(root, info):
        if root.line_type:
            return root.line_type.name

    def resolve_line(root, info):
        if root.line:
            line_object_dict = root.line.__dict__
            line_object_dict.pop('_state', None)
            key_values = list(line_object_dict.items())
            line_object_dict.clear()
            for k, v in key_values:
                new_key = underscore_to_camel(k)
                line_object_dict[new_key] = v
            line_object_dict = json.dumps(line_object_dict, cls=DjangoJSONEncoder)
            return line_object_dict

    class Meta:
        model = BillItem
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            **GenericFilterGQLTypeMixin.get_base_filters_invoice_line_item(),
            **prefix_filterset("bill__", BillGQLType._meta.filter_fields),
        }

        connection_class = ExtendedConnection

        @classmethod
        def get_queryset(cls, queryset, info):
            return BillItem.get_queryset(queryset, info)


class BillPaymentGQLType(DjangoObjectType, GenericFilterGQLTypeMixin):
    class Meta:
        model = BillPayment
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            **GenericFilterGQLTypeMixin.get_base_filters_invoice_payment(),
            **prefix_filterset("bill__", BillGQLType._meta.filter_fields),
        }

        connection_class = ExtendedConnection

        @classmethod
        def get_queryset(cls, queryset, info):
            return BillPayment.get_queryset(queryset, info)


class BillEventGQLType(DjangoObjectType, GenericFilterGQLTypeMixin):
    class Meta:
        model = BillEvent
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            **prefix_filterset("bill__", BillGQLType._meta.filter_fields),
            **GenericFilterGQLTypeMixin.get_base_filters_invoice_event(),
        }

        connection_class = ExtendedConnection

        @classmethod
        def get_queryset(cls, queryset, info):
            return BillEvent.get_queryset(queryset, info)



from datetime import date
from unittest.mock import MagicMock

from core.service_signals import ServiceSignalBindType
from core.signals import REGISTERED_SERVICE_SIGNALS
from invoice.services import InvoiceService
from invoice.tests import DEFAULT_TEST_INVOICE_PAYLOAD
from invoice.tests.gql.base import InvoiceGQLTestCase


class InvoiceGQLTest(InvoiceGQLTestCase):

    create_invoice_mutation = '''
mutation {
  generateInvoicesForTimePeriod(input: {dateFrom: "2021-01-01", dateTo:"2021-10-01"}){
    internalId
    clientMutationId
  }
}    
'''

    search_for_invoice_query = F'''
query {{ 
	invoice(code_Iexact:"{DEFAULT_TEST_INVOICE_PAYLOAD['code']}"){{
    edges {{
      node {{
        isDeleted,
        code,
        codeTp,
        codeExt,
        subjectId,
        subjectType,
        thirdpartyId,
        thirdpartyType,
      }}
    }}
  }}
}}
'''

    def test_mutation_invoice_generate_for_time_frame(self):
        signal_receiver_mock = MagicMock(return_value='123')
        date_from = date(2021, 1, 1)
        date_to = date(2021, 10, 1)
        self.setup_test_signal(signal_receiver_mock)
        _expected_call_args = {
            'signal': REGISTERED_SERVICE_SIGNALS['invoice_creation_from_calculation'].after_service_signal,
            'sender': InvoiceService,
            'cls_': InvoiceService,
            'data': [(), {'user': self.user, 'from_date': date_from, 'to_date': date_to}],
            'context': None,
            'result': None
        }
        mutation = self.create_invoice_mutation
        self.graph_client.execute(mutation, context=self.user_context.get_request())
        signal_receiver_mock.assert_called_once_with(**_expected_call_args)
        pass

    def test_fetch_invoice_query(self):
        output = self.graph_client.execute(self.search_for_invoice_query, context=self.user_context.get_request())
        expected = \
            {'data': {
                'invoice': {
                    'edges': [
                        {'node': {
                            'code': F'{DEFAULT_TEST_INVOICE_PAYLOAD["code"]}',
                            'codeExt': F'{DEFAULT_TEST_INVOICE_PAYLOAD["code_ext"]}',
                            'codeTp': F'{DEFAULT_TEST_INVOICE_PAYLOAD["code_tp"]}',
                            'isDeleted': False,
                            'thirdpartyId': F'{self.invoice.thirdparty.id}',
                            'thirdpartyType': self.invoice.thirdparty_type.id,
                            'subjectId': F'{self.invoice.subject.id}',
                            'subjectType': self.invoice.subject_type.id
        }}]}}}

        self.assertEqual(output, expected)

    def setup_test_signal(self, receiver_mock):
        """
        Mutation doesn't provide logic for generating invoices, just invokes signal.
        """
        from core.signals import bind_service_signal
        bind_service_signal(
            'invoice_creation_from_calculation',
            receiver_mock,
            bind_type=ServiceSignalBindType.AFTER
        )



from decimal import Decimal

from invoice.tests.gql.base import InvoiceGQLTestCase
from invoice.tests.helpers import DEFAULT_TEST_INVOICE_LINE_ITEM_PAYLOAD, DEFAULT_TEST_INVOICE_PAYLOAD


class InvoiceLineItemGQLTest(InvoiceGQLTestCase):

    search_for_invoice_query = F'''
query {{ 
	invoiceLineItem(invoice_Code:"{DEFAULT_TEST_INVOICE_PAYLOAD['code']}", 
	amountTotal:"{DEFAULT_TEST_INVOICE_LINE_ITEM_PAYLOAD['amount_total']}"){{
    edges {{
      node {{
        isDeleted,
        code,
        amountNet,
        amountTotal,
        taxRate,
        lineId,
        lineType
      }}
    }}
  }}
}}
'''

    def test_fetch_invoice_query(self):
        output = self.graph_client.execute(self.search_for_invoice_query, context=self.user_context.get_request())
        expected = \
            {'data': {
                'invoiceLineItem': {
                    'edges': [
                        {'node': {
                            'isDeleted': False,
                            'code': 'LineItem1',
                            'amountNet': "89.50",
                            'amountTotal': "91.50",
                            'taxRate': None,
                            'lineId': F'{self.invoice_line_item.line.id}',
                            'lineType': self.invoice_line_item.line_type.id
                        }}]}}}
        self.assertEqual(output, expected)



from invoice.services.invoice import InvoiceService
from invoice.services.invoiceLineItem import InvoiceLineItemService
from invoice.services.invoicePayments import InvoicePaymentsService
from invoice.services.bill import BillService
from invoice.services.billLineItem import BillLineItemService
from invoice.services.paymentInvoice import PaymentInvoiceService



from graphene import Schema
from graphene.test import Client
from policy.test_helpers import create_test_policy
from product.test_helpers import create_test_product
from core.test_helpers import create_test_interactive_user

from core.forms import User
from django.test import TestCase

from contract.tests.helpers import create_test_contract
from policyholder.tests.helpers import create_test_policy_holder
from insuree.test_helpers import create_test_insuree

from invoice import schema as invoice_schema
from invoice.tests.helpers import create_test_invoice, \
    create_test_invoice_line_item
from invoice.tests.helpers.invoice_payment_helpers import create_test_invoice_payment

from core.models.openimis_graphql_test_case import openIMISGraphQLTestCase, BaseTestContext
class InvoiceGQLTestCase(openIMISGraphQLTestCase):


    @classmethod
    def setUpClass(cls):
        super(InvoiceGQLTestCase, cls).setUpClass()
        cls._graphene_setup()

        cls.maxDiff = None
        cls.user = create_test_interactive_user(username="admin_invoice")
        cls.user_context = BaseTestContext(cls.user)
        cls.policy_holder = create_test_policy_holder()
        cls.contract = create_test_contract(cls.policy_holder)
        cls.insuree = create_test_insuree(with_family=True)
        cls.product = create_test_product("TestC0d3", custom_props={"insurance_period": 12})
        cls.policy = create_test_policy(
            product=cls.product,
            insuree=cls.insuree
        )

        cls.invoice = create_test_invoice(cls.contract, cls.insuree)
        cls.invoice_line_item = \
            create_test_invoice_line_item(invoice=cls.invoice, line_item=cls.policy, user=cls.user)
        cls.invoice_payment = create_test_invoice_payment(invoice=cls.invoice, user=cls.user)

    @classmethod
    def _graphene_setup(cls):
        cls.schema = Schema(
            query=invoice_schema.Query,
            mutation=invoice_schema.Mutation
        )
        cls.graph_client = Client(cls.schema)



# Generated by Django 3.0.14 on 2022-04-11 10:53

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('invoice', '0006_detailpaymentinvoice_historicaldetailpaymentinvoice_historicalpaymentinvoice_paymentinvoice'),
    ]

    operations = [
        migrations.AlterField(
            model_name='detailpaymentinvoice',
            name='subject_type',
            field=models.ForeignKey(db_column='SubjectType', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='subject_type_payment', to='contenttypes.ContentType'),
        ),
    ]



import graphene
import graphene_django_optimizer as gql_optimizer

from django.contrib.auth.models import AnonymousUser
from django.db.models import Q

from core.schema import OrderedDjangoFilterConnectionField
from core.utils import append_validity_filter
from invoice.apps import InvoiceConfig
from invoice.gql.gql_types.payment_types import DetailPaymentInvoiceGQLType
from invoice.models import DetailPaymentInvoice


class DetailPaymentInvoiceQueryMixin:
    detail_payment_invoice = OrderedDjangoFilterConnectionField(
        DetailPaymentInvoiceGQLType,
        orderBy=graphene.List(of_type=graphene.String),
        dateValidFrom__Gte=graphene.DateTime(),
        dateValidTo__Lte=graphene.DateTime(),
        applyDefaultValidityFilter=graphene.Boolean(),
        client_mutation_id=graphene.String(),
    )

    def resolve_detail_payment_invoice(self, info, **kwargs):
        filters = []
        filters += append_validity_filter(**kwargs)

        client_mutation_id = kwargs.get("client_mutation_id", None)
        if client_mutation_id:
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        DetailPaymentInvoiceQueryMixin._check_permissions(info.context.user)
        return gql_optimizer.query(DetailPaymentInvoice.objects.filter(*filters).all(), info)

    @staticmethod
    def _check_permissions(user):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_invoice_payment_search_perms):
            raise PermissionError("Unauthorized")



from django.contrib.contenttypes.models import ContentType
from policy.test_helpers import create_test_policy
from product.models import Product
from product.test_helpers import create_test_product

from insuree.test_helpers import create_test_insuree
from core.forms import User

from invoice.models import Bill, BillItem
from invoice.tests.helpers import create_test_bill
from invoice.tests.helpers.default_test_data import DEFAULT_TEST_BILL_LINE_ITEM_PAYLOAD


def create_test_bill_line_item(bill=None, line_item=None, user=None, **custom_props):
    payload = DEFAULT_TEST_BILL_LINE_ITEM_PAYLOAD.copy()
    payload['bill'] = bill or create_test_bill()
    payload['line'] = line_item or __create_test_policy()
    payload.update(**custom_props)

    BillItem.objects.filter(code=payload['code']).delete()

    user = user or __get_or_create_user()
    invoice = BillItem(**payload)
    invoice.save(username=user.username)

    return invoice


def __get_or_create_user():
    user = User.objects.filter(username='admin_invoice').first()
    if not user:
        user = User.objects.create_superuser(username='admin_invoice', password='S\/pe®Pąßw0rd™')
    return user


def __get_or_create_product():
    product = Product.objects.filter(code='InvLine').first()
    if not product:
        product = create_test_product("InvLine", custom_props={"insurance_period": 12})
    return product


def __create_test_policy():
    insuree = create_test_insuree(with_family=True)
    product = __get_or_create_product()
    return create_test_policy(
        product=product,
        insuree=insuree
    )



import uuid

from unittest import skip

from core.models import MutationLog
from invoice.models import InvoicePayment, InvoicePaymentMutation
from invoice.tests.gql.base import InvoiceGQLTestCase
from invoice.tests.helpers import DEFAULT_TEST_INVOICE_PAYMENT_PAYLOAD, DEFAULT_TEST_INVOICE_PAYLOAD


class InvoicePaymentGQLTest(InvoiceGQLTestCase):

    search_for_invoice_payment_query = F'''
query {{ 
	invoicePayment(invoice_Code:"{DEFAULT_TEST_INVOICE_PAYLOAD['code']}", 
	amountPayed: {DEFAULT_TEST_INVOICE_PAYMENT_PAYLOAD['amount_payed']}){{
    edges {{
      node {{
        isDeleted,
        amountPayed,
        amountReceived,
        invoice {{
            code
        }},
        datePayment,
        status,
        fees
      }}
    }}
  }}
}}
'''

    create_mutation_str = '''  
mutation {{
  createInvoicePayment(input:{{status: 1, codeExt:"{payment_code}", codeTp:"gqlTp", codeReceipt:"gqlRec", 
  label:"gql label", invoiceId:"{invoice_id}", amountPayed: "10.0", fees: "10.0", amountReceived: "20.0", 
  datePayment:"2021-10-10", clientMutationId: "{mutation_id}"}}) {{
    internalId
    clientMutationId
  }}
}}
'''

    delete_mutation_str = '''  
mutation {{
  deleteInvoicePayment(input:{{uuids:["{payment_uuid}"], clientMutationId: "{mutation_id}"}}) {{
    internalId
    clientMutationId
  }}
}}
'''

    update_mutation_str = '''  
mutation {{
	updateInvoicePayment(input:{{id:"{payment_uuid}", codeExt:"updExt", clientMutationId: "{mutation_id}"}}){{
    internalId
    clientMutationId
  }}
}}
'''

    @skip("This is related to old payment approach model - depreceated")
    def test_fetch_invoice_query(self):
        output = self.graph_client.execute(self.search_for_invoice_payment_query,
                                           context=self.BaseTestContext(self.user))
        expected = \
            {'data': {
                'invoicePayment': {
                    'edges': [
                        {'node': {
                            'isDeleted': False,
                            'amountPayed': f"{DEFAULT_TEST_INVOICE_PAYMENT_PAYLOAD['amount_payed']}",
                            'amountReceived': f"{DEFAULT_TEST_INVOICE_PAYMENT_PAYLOAD['amount_received']}",
                            'invoice': {'code': F'{DEFAULT_TEST_INVOICE_PAYLOAD["code"]}'},
                            'datePayment': '2021-10-10',
                            'status': 'A_1',
                            'fees': f"{DEFAULT_TEST_INVOICE_PAYMENT_PAYLOAD['fees']}",
        }}]}}}
        self.assertEqual(output, expected)

    @skip("This is related to old payment approach model - depreceated")
    def test_create_payment_mutation(self):
        payment_code = "GQLCOD"
        mutation_client_id = str(uuid.uuid4())
        mutation = self.create_mutation_str.format(
            payment_code=payment_code, invoice_id=self.invoice.id, mutation_id=mutation_client_id
        )
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))
        expected = InvoicePayment.objects.get(code_ext=payment_code)
        mutation_log = MutationLog.objects.filter(client_mutation_id=mutation_client_id).first()
        obj = InvoicePaymentMutation.objects.get(mutation_id=mutation_log.id).invoice_payment
        self.assertEqual(obj, expected)

    @skip("This is related to old payment approach model - depreceated")
    def test_delete_payment_mutation(self):
        payment_code = "GQLCOD"
        mutation_client_id = str(uuid.uuid4())
        mutation = self.create_mutation_str.format(
            payment_code=payment_code, invoice_id=self.invoice.id, mutation_id=mutation_client_id
        )
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))
        expected = InvoicePayment.objects.get(code_ext=payment_code)
        mutation_client_id = str(uuid.uuid4())
        mutation = self.delete_mutation_str.format(payment_uuid=expected.id, mutation_id=mutation_client_id)
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))
        # TODO: Currently deleted entries are not filtered by manager, only in GQL Query. Should we change this?
        payment = InvoicePayment.objects.filter(code_ext=payment_code).all()
        mutation_ = InvoicePaymentMutation.objects.filter(invoice_payment=payment[0]).all()
        self.assertEqual(len(payment), 1)
        self.assertTrue(payment[0].is_deleted)
        self.assertTrue(len(mutation_) == 2)

    @skip("This is related to old payment approach model - depreceated")
    def test_update_payment_mutation(self):
        payment_code = "GQLCOD"
        mutation_client_id = str(uuid.uuid4())
        mutation = self.create_mutation_str.format(
            payment_code=payment_code, invoice_id=self.invoice.id, mutation_id=mutation_client_id
        )
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))

        created = InvoicePayment.objects.get(code_ext=payment_code)
        mutation_client_id = str(uuid.uuid4())
        mutation = self.update_mutation_str.format(payment_uuid=created.id, mutation_id=mutation_client_id)
        self.graph_client.execute(mutation, context=self.BaseTestContext(self.user))

        expected_code_ext = "updExt"
        mutation_log = MutationLog.objects.filter(client_mutation_id=mutation_client_id).first()
        obj: InvoicePayment = InvoicePaymentMutation.objects.get(mutation_id=mutation_log.id).invoice_payment
        self.assertEqual(obj.code_ext, expected_code_ext)



# Generated by Django 4.2.9 on 2024-02-07 16:47

import core.fields
import datetime
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
        ('invoice', '0012_auto_20230921_0835'),
    ]

    operations = [
        migrations.AlterField(
            model_name='bill',
            name='code_ext',
            field=models.CharField(blank=True, db_column='CodeExt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='bill',
            name='code_tp',
            field=models.CharField(blank=True, db_column='CodeTp', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='bill',
            name='date_bill',
            field=core.fields.DateField(blank=True, db_column='DateBill', default=datetime.date.today, null=True),
        ),
        migrations.AlterField(
            model_name='bill',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='bill',
            name='date_due',
            field=core.fields.DateField(blank=True, db_column='DateDue', null=True),
        ),
        migrations.AlterField(
            model_name='bill',
            name='date_payed',
            field=core.fields.DateField(blank=True, db_column='DatePayed', null=True),
        ),
        migrations.AlterField(
            model_name='bill',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='bill',
            name='payment_reference',
            field=models.CharField(blank=True, db_column='PaymentReference', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='bill',
            name='replacement_uuid',
            field=models.UUIDField(blank=True, db_column='ReplacementUUID', null=True),
        ),
        migrations.AlterField(
            model_name='bill',
            name='subject_id',
            field=models.CharField(blank=True, db_column='SubjectId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='bill',
            name='subject_type',
            field=models.ForeignKey(blank=True, db_column='SubjectType', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='subject_type_bill', to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='bill',
            name='tax_analysis',
            field=models.JSONField(blank=True, db_column='TaxAnalysis', null=True),
        ),
        migrations.AlterField(
            model_name='bill',
            name='thirdparty_id',
            field=models.CharField(blank=True, db_column='ThirdpartyId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='bill',
            name='thirdparty_type',
            field=models.ForeignKey(blank=True, db_column='ThirdpartyType', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='bill',
            name='user_created',
            field=models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_created', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='bill',
            name='user_updated',
            field=models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_updated', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='billevent',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='billevent',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='billevent',
            name='message',
            field=models.CharField(blank=True, db_column='Message', max_length=500, null=True),
        ),
        migrations.AlterField(
            model_name='billevent',
            name='user_created',
            field=models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_created', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='billevent',
            name='user_updated',
            field=models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_updated', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='billitem',
            name='amount_total',
            field=models.DecimalField(db_column='AmountTotal', decimal_places=2, default=0.0, max_digits=18),
        ),
        migrations.AlterField(
            model_name='billitem',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='billitem',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='billitem',
            name='details',
            field=models.JSONField(blank=True, db_column='Details', null=True),
        ),
        migrations.AlterField(
            model_name='billitem',
            name='ledger_account',
            field=models.CharField(blank=True, db_column='LedgerAccount', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='billitem',
            name='line_id',
            field=models.CharField(blank=True, db_column='LineId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='billitem',
            name='line_type',
            field=models.ForeignKey(blank=True, db_column='LineType', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='line_type_bill', to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='billitem',
            name='replacement_uuid',
            field=models.UUIDField(blank=True, db_column='ReplacementUUID', null=True),
        ),
        migrations.AlterField(
            model_name='billitem',
            name='tax_analysis',
            field=models.JSONField(blank=True, db_column='TaxAnalysis', null=True),
        ),
        migrations.AlterField(
            model_name='billitem',
            name='tax_rate',
            field=models.UUIDField(blank=True, db_column='CalculationUUID', null=True),
        ),
        migrations.AlterField(
            model_name='billitem',
            name='user_created',
            field=models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_created', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='billitem',
            name='user_updated',
            field=models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_updated', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='billpayment',
            name='amount_payed',
            field=models.DecimalField(blank=True, db_column='AmountPayed', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='billpayment',
            name='amount_received',
            field=models.DecimalField(blank=True, db_column='AmountReceived', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='billpayment',
            name='code_ext',
            field=models.CharField(blank=True, db_column='CodeExt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='billpayment',
            name='code_receipt',
            field=models.CharField(blank=True, db_column='CodeReceipt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='billpayment',
            name='code_tp',
            field=models.CharField(blank=True, db_column='CodeTp', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='billpayment',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='billpayment',
            name='date_payment',
            field=core.fields.DateField(blank=True, db_column='DatePayment', null=True),
        ),
        migrations.AlterField(
            model_name='billpayment',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='billpayment',
            name='fees',
            field=models.DecimalField(blank=True, db_column='Fees', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='billpayment',
            name='label',
            field=models.CharField(blank=True, db_column='Label', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='billpayment',
            name='payment_origin',
            field=models.CharField(blank=True, db_column='PaymentOrigin', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='billpayment',
            name='user_created',
            field=models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_created', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='billpayment',
            name='user_updated',
            field=models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_updated', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='detailpaymentinvoice',
            name='amount',
            field=models.DecimalField(blank=True, db_column='Amount', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='detailpaymentinvoice',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='detailpaymentinvoice',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='detailpaymentinvoice',
            name='fees',
            field=models.DecimalField(blank=True, db_column='Fees', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='detailpaymentinvoice',
            name='reconcilation_date',
            field=models.DateField(blank=True, db_column='ReconcilationDate', null=True),
        ),
        migrations.AlterField(
            model_name='detailpaymentinvoice',
            name='reconcilation_id',
            field=models.CharField(blank=True, db_column='ReconcilationId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='detailpaymentinvoice',
            name='subject_id',
            field=models.CharField(blank=True, db_column='SubjectId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='detailpaymentinvoice',
            name='subject_type',
            field=models.ForeignKey(blank=True, db_column='SubjectType', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='subject_type_payment', to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='detailpaymentinvoice',
            name='user_created',
            field=models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_created', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='detailpaymentinvoice',
            name='user_updated',
            field=models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_updated', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='historicalbill',
            name='code_ext',
            field=models.CharField(blank=True, db_column='CodeExt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbill',
            name='code_tp',
            field=models.CharField(blank=True, db_column='CodeTp', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbill',
            name='date_bill',
            field=core.fields.DateField(blank=True, db_column='DateBill', default=datetime.date.today, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbill',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbill',
            name='date_due',
            field=core.fields.DateField(blank=True, db_column='DateDue', null=True),
        ),
        migrations.AlterField(
            model_name='historicalbill',
            name='date_payed',
            field=core.fields.DateField(blank=True, db_column='DatePayed', null=True),
        ),
        migrations.AlterField(
            model_name='historicalbill',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbill',
            name='payment_reference',
            field=models.CharField(blank=True, db_column='PaymentReference', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbill',
            name='replacement_uuid',
            field=models.UUIDField(blank=True, db_column='ReplacementUUID', null=True),
        ),
        migrations.AlterField(
            model_name='historicalbill',
            name='subject_id',
            field=models.CharField(blank=True, db_column='SubjectId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbill',
            name='tax_analysis',
            field=models.JSONField(blank=True, db_column='TaxAnalysis', null=True),
        ),
        migrations.AlterField(
            model_name='historicalbill',
            name='thirdparty_id',
            field=models.CharField(blank=True, db_column='ThirdpartyId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillevent',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillevent',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillevent',
            name='message',
            field=models.CharField(blank=True, db_column='Message', max_length=500, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillitem',
            name='amount_total',
            field=models.DecimalField(db_column='AmountTotal', decimal_places=2, default=0.0, max_digits=18),
        ),
        migrations.AlterField(
            model_name='historicalbillitem',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillitem',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillitem',
            name='details',
            field=models.JSONField(blank=True, db_column='Details', null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillitem',
            name='ledger_account',
            field=models.CharField(blank=True, db_column='LedgerAccount', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillitem',
            name='line_id',
            field=models.CharField(blank=True, db_column='LineId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillitem',
            name='replacement_uuid',
            field=models.UUIDField(blank=True, db_column='ReplacementUUID', null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillitem',
            name='tax_analysis',
            field=models.JSONField(blank=True, db_column='TaxAnalysis', null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillitem',
            name='tax_rate',
            field=models.UUIDField(blank=True, db_column='CalculationUUID', null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillpayment',
            name='amount_payed',
            field=models.DecimalField(blank=True, db_column='AmountPayed', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillpayment',
            name='amount_received',
            field=models.DecimalField(blank=True, db_column='AmountReceived', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillpayment',
            name='code_ext',
            field=models.CharField(blank=True, db_column='CodeExt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillpayment',
            name='code_receipt',
            field=models.CharField(blank=True, db_column='CodeReceipt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillpayment',
            name='code_tp',
            field=models.CharField(blank=True, db_column='CodeTp', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillpayment',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillpayment',
            name='date_payment',
            field=core.fields.DateField(blank=True, db_column='DatePayment', null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillpayment',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillpayment',
            name='fees',
            field=models.DecimalField(blank=True, db_column='Fees', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillpayment',
            name='label',
            field=models.CharField(blank=True, db_column='Label', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbillpayment',
            name='payment_origin',
            field=models.CharField(blank=True, db_column='PaymentOrigin', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicaldetailpaymentinvoice',
            name='amount',
            field=models.DecimalField(blank=True, db_column='Amount', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='historicaldetailpaymentinvoice',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicaldetailpaymentinvoice',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicaldetailpaymentinvoice',
            name='fees',
            field=models.DecimalField(blank=True, db_column='Fees', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='historicaldetailpaymentinvoice',
            name='reconcilation_date',
            field=models.DateField(blank=True, db_column='ReconcilationDate', null=True),
        ),
        migrations.AlterField(
            model_name='historicaldetailpaymentinvoice',
            name='reconcilation_id',
            field=models.CharField(blank=True, db_column='ReconcilationId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicaldetailpaymentinvoice',
            name='subject_id',
            field=models.CharField(blank=True, db_column='SubjectId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoice',
            name='code_ext',
            field=models.CharField(blank=True, db_column='CodeExt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoice',
            name='code_tp',
            field=models.CharField(blank=True, db_column='CodeTp', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoice',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoice',
            name='date_due',
            field=core.fields.DateField(blank=True, db_column='DateDue', null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoice',
            name='date_invoice',
            field=core.fields.DateField(blank=True, db_column='DateInvoice', default=datetime.date.today, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoice',
            name='date_payed',
            field=core.fields.DateField(blank=True, db_column='DatePayed', null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoice',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoice',
            name='payment_reference',
            field=models.CharField(blank=True, db_column='PaymentReference', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoice',
            name='replacement_uuid',
            field=models.UUIDField(blank=True, db_column='ReplacementUUID', null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoice',
            name='subject_id',
            field=models.CharField(blank=True, db_column='SubjectId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoice',
            name='tax_analysis',
            field=models.JSONField(blank=True, db_column='TaxAnalysis', null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoice',
            name='thirdparty_id',
            field=models.CharField(blank=True, db_column='ThirdpartyId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoiceevent',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoiceevent',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoiceevent',
            name='message',
            field=models.CharField(blank=True, db_column='Message', max_length=500, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicelineitem',
            name='amount_total',
            field=models.DecimalField(db_column='AmountTotal', decimal_places=2, default=0.0, max_digits=18),
        ),
        migrations.AlterField(
            model_name='historicalinvoicelineitem',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicelineitem',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicelineitem',
            name='details',
            field=models.JSONField(blank=True, db_column='Details', null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicelineitem',
            name='ledger_account',
            field=models.CharField(blank=True, db_column='LedgerAccount', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicelineitem',
            name='line_id',
            field=models.CharField(blank=True, db_column='LineId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicelineitem',
            name='replacement_uuid',
            field=models.UUIDField(blank=True, db_column='ReplacementUUID', null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicelineitem',
            name='tax_analysis',
            field=models.JSONField(blank=True, db_column='TaxAnalysis', null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicelineitem',
            name='tax_rate',
            field=models.UUIDField(blank=True, db_column='CalculationUUID', null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicepayment',
            name='amount_payed',
            field=models.DecimalField(blank=True, db_column='AmountPayed', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicepayment',
            name='amount_received',
            field=models.DecimalField(blank=True, db_column='AmountReceived', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicepayment',
            name='code_ext',
            field=models.CharField(blank=True, db_column='CodeExt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicepayment',
            name='code_receipt',
            field=models.CharField(blank=True, db_column='CodeReceipt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicepayment',
            name='code_tp',
            field=models.CharField(blank=True, db_column='CodeTp', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicepayment',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicepayment',
            name='date_payment',
            field=core.fields.DateField(blank=True, db_column='DatePayment', null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicepayment',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicepayment',
            name='fees',
            field=models.DecimalField(blank=True, db_column='Fees', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicepayment',
            name='label',
            field=models.CharField(blank=True, db_column='Label', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalinvoicepayment',
            name='payment_origin',
            field=models.CharField(blank=True, db_column='PaymentOrigin', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalpaymentinvoice',
            name='amount_received',
            field=models.DecimalField(blank=True, db_column='AmountReceived', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='historicalpaymentinvoice',
            name='code_ext',
            field=models.CharField(blank=True, db_column='CodeExt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalpaymentinvoice',
            name='code_receipt',
            field=models.CharField(blank=True, db_column='CodeReceipt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalpaymentinvoice',
            name='code_tp',
            field=models.CharField(blank=True, db_column='CodeTp', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalpaymentinvoice',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalpaymentinvoice',
            name='date_payment',
            field=core.fields.DateField(blank=True, db_column='DatePayment', null=True),
        ),
        migrations.AlterField(
            model_name='historicalpaymentinvoice',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='historicalpaymentinvoice',
            name='fees',
            field=models.DecimalField(blank=True, db_column='Fees', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='historicalpaymentinvoice',
            name='label',
            field=models.CharField(blank=True, db_column='Label', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalpaymentinvoice',
            name='payer_name',
            field=models.CharField(blank=True, db_column='PayerName', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalpaymentinvoice',
            name='payment_origin',
            field=models.CharField(blank=True, db_column='PaymentOrigin', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='code_ext',
            field=models.CharField(blank=True, db_column='CodeExt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='code_tp',
            field=models.CharField(blank=True, db_column='CodeTp', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='date_due',
            field=core.fields.DateField(blank=True, db_column='DateDue', null=True),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='date_invoice',
            field=core.fields.DateField(blank=True, db_column='DateInvoice', default=datetime.date.today, null=True),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='date_payed',
            field=core.fields.DateField(blank=True, db_column='DatePayed', null=True),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='payment_reference',
            field=models.CharField(blank=True, db_column='PaymentReference', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='replacement_uuid',
            field=models.UUIDField(blank=True, db_column='ReplacementUUID', null=True),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='subject_id',
            field=models.CharField(blank=True, db_column='SubjectId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='subject_type',
            field=models.ForeignKey(blank=True, db_column='SubjectType', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='subject_type', to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='tax_analysis',
            field=models.JSONField(blank=True, db_column='TaxAnalysis', null=True),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='thirdparty_id',
            field=models.CharField(blank=True, db_column='ThirdpartyId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='thirdparty_type',
            field=models.ForeignKey(blank=True, db_column='ThirdpartyType', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='user_created',
            field=models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_created', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='user_updated',
            field=models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_updated', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='invoiceevent',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='invoiceevent',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='invoiceevent',
            name='message',
            field=models.CharField(blank=True, db_column='Message', max_length=500, null=True),
        ),
        migrations.AlterField(
            model_name='invoiceevent',
            name='user_created',
            field=models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_created', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='invoiceevent',
            name='user_updated',
            field=models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_updated', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='invoicelineitem',
            name='amount_total',
            field=models.DecimalField(db_column='AmountTotal', decimal_places=2, default=0.0, max_digits=18),
        ),
        migrations.AlterField(
            model_name='invoicelineitem',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='invoicelineitem',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='invoicelineitem',
            name='details',
            field=models.JSONField(blank=True, db_column='Details', null=True),
        ),
        migrations.AlterField(
            model_name='invoicelineitem',
            name='ledger_account',
            field=models.CharField(blank=True, db_column='LedgerAccount', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='invoicelineitem',
            name='line_id',
            field=models.CharField(blank=True, db_column='LineId', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='invoicelineitem',
            name='line_type',
            field=models.ForeignKey(blank=True, db_column='LineType', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='line_type', to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='invoicelineitem',
            name='replacement_uuid',
            field=models.UUIDField(blank=True, db_column='ReplacementUUID', null=True),
        ),
        migrations.AlterField(
            model_name='invoicelineitem',
            name='tax_analysis',
            field=models.JSONField(blank=True, db_column='TaxAnalysis', null=True),
        ),
        migrations.AlterField(
            model_name='invoicelineitem',
            name='tax_rate',
            field=models.UUIDField(blank=True, db_column='CalculationUUID', null=True),
        ),
        migrations.AlterField(
            model_name='invoicelineitem',
            name='user_created',
            field=models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_created', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='invoicelineitem',
            name='user_updated',
            field=models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_updated', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='invoicepayment',
            name='amount_payed',
            field=models.DecimalField(blank=True, db_column='AmountPayed', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='invoicepayment',
            name='amount_received',
            field=models.DecimalField(blank=True, db_column='AmountReceived', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='invoicepayment',
            name='code_ext',
            field=models.CharField(blank=True, db_column='CodeExt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='invoicepayment',
            name='code_receipt',
            field=models.CharField(blank=True, db_column='CodeReceipt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='invoicepayment',
            name='code_tp',
            field=models.CharField(blank=True, db_column='CodeTp', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='invoicepayment',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='invoicepayment',
            name='date_payment',
            field=core.fields.DateField(blank=True, db_column='DatePayment', null=True),
        ),
        migrations.AlterField(
            model_name='invoicepayment',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='invoicepayment',
            name='fees',
            field=models.DecimalField(blank=True, db_column='Fees', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='invoicepayment',
            name='label',
            field=models.CharField(blank=True, db_column='Label', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='invoicepayment',
            name='payment_origin',
            field=models.CharField(blank=True, db_column='PaymentOrigin', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='invoicepayment',
            name='user_created',
            field=models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_created', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='invoicepayment',
            name='user_updated',
            field=models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_updated', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='paymentinvoice',
            name='amount_received',
            field=models.DecimalField(blank=True, db_column='AmountReceived', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='paymentinvoice',
            name='code_ext',
            field=models.CharField(blank=True, db_column='CodeExt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='paymentinvoice',
            name='code_receipt',
            field=models.CharField(blank=True, db_column='CodeReceipt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='paymentinvoice',
            name='code_tp',
            field=models.CharField(blank=True, db_column='CodeTp', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='paymentinvoice',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='paymentinvoice',
            name='date_payment',
            field=core.fields.DateField(blank=True, db_column='DatePayment', null=True),
        ),
        migrations.AlterField(
            model_name='paymentinvoice',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', default=datetime.datetime.now, null=True),
        ),
        migrations.AlterField(
            model_name='paymentinvoice',
            name='fees',
            field=models.DecimalField(blank=True, db_column='Fees', decimal_places=2, max_digits=18, null=True),
        ),
        migrations.AlterField(
            model_name='paymentinvoice',
            name='label',
            field=models.CharField(blank=True, db_column='Label', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='paymentinvoice',
            name='payer_name',
            field=models.CharField(blank=True, db_column='PayerName', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='paymentinvoice',
            name='payment_origin',
            field=models.CharField(blank=True, db_column='PaymentOrigin', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='paymentinvoice',
            name='user_created',
            field=models.ForeignKey(db_column='UserCreatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_created', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='paymentinvoice',
            name='user_updated',
            field=models.ForeignKey(db_column='UserUpdatedUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='%(class)s_user_updated', to=settings.AUTH_USER_MODEL),
        ),
    ]



import graphene
from django.contrib.auth.models import AnonymousUser
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError

from core.gql.gql_mutations.base_mutation import (
    BaseMutation,
    BaseHistoryModelCreateMutationMixin,
    BaseHistoryModelUpdateMutationMixin,
    BaseHistoryModelDeleteMutationMixin
)
from core.schema import OpenIMISMutation
from invoice.apps import InvoiceConfig
from invoice.gql.input_types import CreateDetailPaymentInvoiceInputType, UpdateDetailPaymentInvoiceInputType
from invoice.models import DetailPaymentInvoice, DetailPaymentInvoiceMutation


class CreateDetailPaymentInvoiceMutation(BaseHistoryModelCreateMutationMixin, BaseMutation):
    _mutation_class = "CreateDetailPaymentInvoiceMutation"
    _mutation_module = "invoice"
    _model = DetailPaymentInvoice

    @classmethod
    def _mutate(cls, user, **data):
        client_mutation_id = data.get("client_mutation_id")
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        data = cls._convert_content_type(data=data)
        detail_payment_invoice = cls.create_object(user=user, object_data=data)
        if detail_payment_invoice:
            DetailPaymentInvoiceMutation.object_mutated(
                user,
                client_mutation_id=client_mutation_id,
                detail_payment_invoice=detail_payment_invoice
            )

    @classmethod
    def _convert_content_type(cls, data):
        if "subject_type" in data:
            subject_type = data.pop('subject_type')
            data['subject_type'] = ContentType.objects.get(model__iexact=subject_type)
        return data

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_invoice_payment_create_perms):
            raise ValidationError("mutation.authentication_required")

    class Input(CreateDetailPaymentInvoiceInputType):
        pass


class UpdateDetailPaymentInvoiceMutation(BaseHistoryModelUpdateMutationMixin, BaseMutation):
    _mutation_class = "UpdateDetailPaymentInvoiceMutation"
    _mutation_module = "invoice"
    _model = DetailPaymentInvoice

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_invoice_payment_update_perms):
            raise ValidationError("mutation.authentication_required")

    class Input(UpdateDetailPaymentInvoiceInputType):
        pass


class DeleteDetailPaymentInvoiceMutation(BaseHistoryModelDeleteMutationMixin, BaseMutation):
    _mutation_class = "DeleteDetailPaymentInvoiceMutation"
    _mutation_module = "invoice"
    _model = DetailPaymentInvoice

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_invoice_payment_delete_perms):
            raise ValidationError("mutation.authentication_required")

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.UUID)



from django.db import models
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType

from core.models import HistoryBusinessModel, HistoryModel, UUIDModel, ObjectMutation, MutationLog
from core.fields import DateField
from datetime import date
from invoice.apps import InvoiceConfig
from django.utils.translation import gettext as _
# Create your models here.
from invoice.mixins import GenericInvoiceQuerysetMixin, GenericInvoiceManager


def get_default_currency():
    return InvoiceConfig.default_currency_code


class GenericInvoice(GenericInvoiceQuerysetMixin, HistoryBusinessModel):
    class Status(models.IntegerChoices):
        DRAFT = 0, _('draft')
        VALIDATED = 1, _('validated')
        PAID = 2, _('paid')
        CANCELLED = 3, _('cancelled')
        DELETED = 4, _('deleted')
        SUSPENDED = 5, _('suspended')
        UNPAID = 6, _('unpaid')
        RECONCILIATED = 7, _('reconciliated')

    thirdparty_type = models.ForeignKey(ContentType, models.DO_NOTHING,
                                          db_column='ThirdpartyType', blank=True, null=True, unique=False)
    thirdparty_id = models.CharField(db_column='ThirdpartyId', max_length=255, blank=True, null=True)  # object is referenced by uuid
    thirdparty = GenericForeignKey('thirdparty_type', 'thirdparty_id')

    code_tp = models.CharField(db_column='CodeTp', max_length=255, blank=True, null=True)
    code = models.CharField(db_column='Code', max_length=255, null=False)
    code_ext = models.CharField(db_column='CodeExt', max_length=255, blank=True, null=True)

    date_due = DateField(db_column='DateDue', blank=True, null=True)

    date_payed = DateField(db_column='DatePayed', blank=True, null=True)

    amount_discount = models.DecimalField(
        db_column='AmountDiscount', max_digits=18, decimal_places=2,  null=True, default=0.0)
    amount_net = models.DecimalField(
        db_column='AmountNet', max_digits=18, decimal_places=2, default=0.0)
    amount_total = models.DecimalField(
        db_column='AmountTotal', max_digits=18, decimal_places=2, default=0.0)

    tax_analysis = models.JSONField(db_column='TaxAnalysis', blank=True, null=True)

    status = models.SmallIntegerField(
        db_column='Status', null=False, choices=Status.choices, default=Status.DRAFT)

    currency_tp_code = models.CharField(
        db_column='CurrencyTpCode', null=False, max_length=255, default=get_default_currency)
    currency_code = models.CharField(
        db_column='CurrencyCode', null=False, max_length=255, default=get_default_currency)

    note = models.TextField(db_column='Note', blank=True, null=True)
    terms = models.TextField(db_column='Terms', blank=True, null=True)

    payment_reference = models.CharField(db_column='PaymentReference', max_length=255, blank=True, null=True)

    objects = GenericInvoiceManager()

    class Meta:
        abstract = True


class GenericInvoiceLineItem(GenericInvoiceQuerysetMixin, HistoryBusinessModel):
    code = models.CharField(db_column='Code', max_length=255, null=False)

    description = models.TextField(db_column='Description', blank=True, null=True)
    details = models.JSONField(db_column='Details', blank=True, null=True)

    ledger_account = models.CharField(db_column='LedgerAccount', max_length=255, blank=True, null=True)

    quantity = models.IntegerField(db_column='Quantity', default=0.0)
    unit_price = models.DecimalField(db_column='UnitPrice', max_digits=18, decimal_places=2, default=0.0)

    discount = models.DecimalField(db_column='Discount', max_digits=18, decimal_places=2, default=0.0)

    deduction = models.DecimalField(db_column='Deduction', max_digits=18, decimal_places=2, default=0.0)

    tax_rate = models.UUIDField(db_column="CalculationUUID", blank=True, null=True)
    tax_analysis = models.JSONField(db_column='TaxAnalysis',  blank=True, null=True)

    amount_total = models.DecimalField(db_column='AmountTotal', max_digits=18, decimal_places=2, default=0.0)
    amount_net = models.DecimalField(db_column='AmountNet', max_digits=18, decimal_places=2, default=0.0)

    objects = GenericInvoiceManager()

    class Meta:
        abstract = True


class GenericInvoicePayment(GenericInvoiceQuerysetMixin, HistoryModel):
    class PaymentStatus(models.IntegerChoices):
        REJECTED = 0, _('rejected')
        ACCEPTED = 1, _('accepted')
        REFUNDED = 2, _('refunded')
        CANCELLED = 3, _('cancelled')

    code_tp = models.CharField(db_column='CodeTp', max_length=255, blank=True, null=True)
    code_ext = models.CharField(db_column='CodeExt', max_length=255, blank=True, null=True)
    code_receipt = models.CharField(db_column='CodeReceipt', max_length=255, blank=True, null=True)

    label = models.CharField(db_column='Label', max_length=255,  blank=True, null=True)

    status = models.SmallIntegerField(db_column='Status', null=False, choices=PaymentStatus.choices)

    amount_payed = models.DecimalField(db_column='AmountPayed', max_digits=18, decimal_places=2,  blank=True, null=True)
    fees = models.DecimalField(db_column='Fees', max_digits=18, decimal_places=2,  blank=True, null=True)
    amount_received = models.DecimalField(db_column='AmountReceived', max_digits=18, decimal_places=2,  blank=True, null=True)

    date_payment = DateField(db_column='DatePayment',  blank=True, null=True)

    payment_origin = models.CharField(db_column='PaymentOrigin', max_length=255,  blank=True, null=True)

    objects = GenericInvoiceManager()

    class Meta:
        abstract = True


class GenericInvoiceEvent(GenericInvoiceQuerysetMixin, HistoryModel):
    class EventType(models.IntegerChoices):
        MESSAGE = 0, _('message')
        STATUS = 1, _('status')
        WARNING = 2, _('warning')
        PAYMENT = 3, _('payment')
        PAYMENT_ERROR = 4, _('payment_error')

    message = models.CharField(db_column='Message', max_length=500,  blank=True, null=True)
    event_type = models.SmallIntegerField(
        db_column='Status', null=False, choices=EventType.choices, default=EventType.MESSAGE)

    objects = GenericInvoiceManager()

    class Meta:
        abstract = True


class Invoice(GenericInvoice):
    subject_type = models.ForeignKey(ContentType, models.DO_NOTHING,
                                        db_column='SubjectType',  blank=True, null=True, related_name='subject_type', unique=False)
    subject_id = models.CharField(db_column='SubjectId', max_length=255,  blank=True, null=True)  # object is referenced by uuid
    subject = GenericForeignKey('subject_type', 'subject_id')

    date_invoice = DateField(db_column='DateInvoice', default=date.today,  blank=True, null=True)

    class Meta:
        managed = True
        db_table = 'tblInvoice'


class InvoiceLineItem(GenericInvoiceLineItem):
    line_type = models.ForeignKey(
        ContentType, models.DO_NOTHING, db_column='LineType',  blank=True, null=True, related_name='line_type', unique=False)
    line_id = models.CharField(db_column='LineId', max_length=255,  blank=True, null=True)  # object is referenced by uuid
    line = GenericForeignKey('line_type', 'line_id')

    invoice = models.ForeignKey(Invoice, models.DO_NOTHING, db_column='InvoiceId', related_name="line_items")

    class Meta:
        managed = True
        db_table = 'tblInvoiceLineItem'


class InvoicePayment(GenericInvoicePayment):
    invoice = models.ForeignKey(Invoice, models.DO_NOTHING, db_column='InvoiceId', related_name="payments")

    class Meta:
        managed = True
        db_table = 'tblInvoicePayment'


class InvoiceEvent(GenericInvoiceEvent):
    invoice = models.ForeignKey(Invoice, models.DO_NOTHING, db_column='InvoiceId', related_name="events")

    class Meta:
        managed = True
        db_table = 'tblInvoiceEvent'


class Bill(GenericInvoice):
    subject_type = models.ForeignKey(ContentType, models.DO_NOTHING,
                                        db_column='SubjectType', null=True,blank=True, related_name='subject_type_bill',
                                     unique=False)
    subject_id = models.CharField(db_column='SubjectId', max_length=255, blank=True, null=True)  # object is referenced by uuid
    subject = GenericForeignKey('subject_type', 'subject_id')

    date_bill = DateField(db_column='DateBill', default=date.today,blank=True, null=True)

    class Meta:
        managed = True
        db_table = 'tblBill'


class BillItem(GenericInvoiceLineItem):
    line_type = models.ForeignKey(
        ContentType, models.DO_NOTHING, db_column='LineType',  blank=True, null=True, related_name='line_type_bill', unique=False)
    line_id = models.CharField(db_column='LineId', max_length=255,  blank=True, null=True)  # object is referenced by uuid
    line = GenericForeignKey('line_type', 'line_id')

    bill = models.ForeignKey(Bill, models.DO_NOTHING, db_column='BillId', related_name="line_items_bill")

    class Meta:
        managed = True
        db_table = 'tblBillLineItem'


class BillPayment(GenericInvoicePayment):
    bill = models.ForeignKey(Bill, models.DO_NOTHING, db_column='BillId', related_name="payments_bill")

    class Meta:
        managed = True
        db_table = 'tblBillPayment'


class BillEvent(GenericInvoiceEvent):
    bill = models.ForeignKey(Bill, models.DO_NOTHING, db_column='BillId', related_name="events_bill")

    class Meta:
        managed = True
        db_table = 'tblBillEvent'


class InvoiceMutation(UUIDModel, ObjectMutation):
    invoice = models.ForeignKey(Invoice, models.DO_NOTHING, related_name='mutations')
    mutation = models.ForeignKey(MutationLog, models.DO_NOTHING, related_name='invoices')

    class Meta:
        managed = True
        db_table = "invoice_invoiceMutation"


class InvoicePaymentMutation(UUIDModel, ObjectMutation):
    invoice_payment = models.ForeignKey(InvoicePayment, models.DO_NOTHING, related_name='mutations')
    mutation = models.ForeignKey(MutationLog, models.DO_NOTHING, related_name='invoice_payments')

    class Meta:
        managed = True
        db_table = "invoice_InvoicePaymentMutation"


class InvoiceLineItemMutation(UUIDModel, ObjectMutation):
    invoice_line_items = models.ForeignKey(InvoiceLineItem, models.DO_NOTHING, related_name='mutations')
    mutation = models.ForeignKey(MutationLog, models.DO_NOTHING, related_name='invoice_line_items')

    class Meta:
        managed = True
        db_table = "invoice_InvoiceLineItemsMutation"


class InvoiceEventMutation(UUIDModel, ObjectMutation):
    invoice_event = models.ForeignKey(InvoiceEvent, models.DO_NOTHING, related_name='mutations')
    mutation = models.ForeignKey(MutationLog, models.DO_NOTHING, related_name='event_messages')

    class Meta:
        managed = True
        db_table = "invoice_InvoiceEventMutation"


class BillMutation(UUIDModel, ObjectMutation):
    bill = models.ForeignKey(Bill, models.DO_NOTHING, related_name='mutations')
    mutation = models.ForeignKey(MutationLog, models.DO_NOTHING, related_name='bills')

    class Meta:
        managed = True
        db_table = "bill_BillMutation"


class BillPaymentMutation(UUIDModel, ObjectMutation):
    bill_payment = models.ForeignKey(BillPayment, models.DO_NOTHING, related_name='mutations')
    mutation = models.ForeignKey(MutationLog, models.DO_NOTHING, related_name='bill_payments')

    class Meta:
        managed = True
        db_table = "bill_BillPaymentMutation"


class BillItemMutation(UUIDModel, ObjectMutation):
    bile_items = models.ForeignKey(BillItem, models.DO_NOTHING, related_name='mutations')
    mutation = models.ForeignKey(MutationLog, models.DO_NOTHING, related_name='bill_line_items')

    class Meta:
        managed = True
        db_table = "bill_BillLineItemsMutation"


class BillEventMutation(UUIDModel, ObjectMutation):
    bill_event = models.ForeignKey(BillEvent, models.DO_NOTHING, related_name='mutations')
    mutation = models.ForeignKey(MutationLog, models.DO_NOTHING, related_name='bill_event_messages')

    class Meta:
        managed = True
        db_table = "bill_BillEventMutation"


# new approach for payment tables: 'PaymentInvoice' and 'DetailPaymentInvoice'
class PaymentInvoice(GenericInvoiceQuerysetMixin, HistoryModel):
    class ReconciliationStatus(models.IntegerChoices):
        NOT_RECONCILIATED = 0, _('not reconciliated')
        RECONCILIATED = 1, _('reconciliated')
        REFUNDED = 2, _('refunded')
        CANCELLED = 3, _('cancelled')

    code_tp = models.CharField(db_column='CodeTp', max_length=255,  blank=True, null=True)
    code_ext = models.CharField(db_column='CodeExt', max_length=255,  blank=True, null=True)
    code_receipt = models.CharField(db_column='CodeReceipt', max_length=255,  blank=True, null=True)

    label = models.CharField(db_column='Label', max_length=255,  blank=True, null=True)

    reconciliation_status = models.SmallIntegerField(db_column='ReconciliationStatus', null=False,
                                                     choices=ReconciliationStatus.choices)

    fees = models.DecimalField(db_column='Fees', max_digits=18, decimal_places=2,  blank=True, null=True)
    amount_received = models.DecimalField(db_column='AmountReceived', max_digits=18, decimal_places=2,  blank=True, null=True)

    date_payment = DateField(db_column='DatePayment',  blank=True, null=True)

    payment_origin = models.CharField(db_column='PaymentOrigin', max_length=255,  blank=True, null=True)

    payer_ref = models.CharField(db_column='PayerRef', max_length=255)
    payer_name = models.CharField(db_column='PayerName', max_length=255,  blank=True, null=True)

    objects = GenericInvoiceManager()

    class Meta:
        managed = True
        db_table = "tblPaymentInvoice"


class DetailPaymentInvoice(GenericInvoiceQuerysetMixin, HistoryModel):
    class DetailPaymentStatus(models.IntegerChoices):
        REJECTED = 0, _('rejected')
        ACCEPTED = 1, _('accepted')
        REFUNDED = 2, _('refunded')
        CANCELLED = 3, _('cancelled')

    payment = models.ForeignKey(PaymentInvoice, models.DO_NOTHING,
                                db_column='PaymentUUID', related_name="invoice_payments")

    subject_type = models.ForeignKey(ContentType, models.DO_NOTHING, db_column='SubjectType',
                                     related_name='subject_type_payment',  blank=True, null=True, unique=False)
    subject_id = models.CharField(db_column='SubjectId', max_length=255,  blank=True, null=True)
    subject = GenericForeignKey('subject_type', 'subject_id')

    status = models.SmallIntegerField(db_column='Status', null=False, choices=DetailPaymentStatus.choices)
    fees = models.DecimalField(db_column='Fees', max_digits=18, decimal_places=2,  blank=True, null=True)
    amount = models.DecimalField(db_column='Amount', max_digits=18, decimal_places=2,  blank=True, null=True)

    reconcilation_id = models.CharField(db_column='ReconcilationId', max_length=255,  blank=True, null=True)
    reconcilation_date = models.DateField(db_column='ReconcilationDate',  blank=True, null=True)

    objects = GenericInvoiceManager()

    class Meta:
        managed = True
        db_table = "tblDetailPaymentInvoice"


class PaymentInvoiceMutation(UUIDModel, ObjectMutation):
    payment_invoice = models.ForeignKey(PaymentInvoice, models.DO_NOTHING, related_name='mutations')
    mutation = models.ForeignKey(MutationLog, models.DO_NOTHING, related_name='payment_invoices')

    class Meta:
        managed = True
        db_table = "paymentinvoice_PaymentInvoiceMutation"


class DetailPaymentInvoiceMutation(UUIDModel, ObjectMutation):
    detail_payment_invoice = models.ForeignKey(DetailPaymentInvoice, models.DO_NOTHING, related_name='mutations')
    mutation = models.ForeignKey(MutationLog, models.DO_NOTHING, related_name='detail_payment_invoices')

    class Meta:
        managed = True
        db_table = "paymentinvoice_DetailPaymentInvoiceMutation"



from abc import ABC

from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _

from core.validation import BaseModelValidation, UniqueCodeValidationMixin, ObjectExistsValidationMixin


class TaxAnalysisFormatValidationMixin:
    INVALID_TAX_ANALYSIS_JSON_FORMAT_NO_KEYS = _("Invalid tax_analysis format, 'total' and 'lines' keys are required.")

    INVALID_TAX_ANALYSIS_JSON_FORMAT_LINES_KEYS = _("tax_analysis.lines is not in list format.")

    INVALID_TAX_ANALYSIS_JSON_FORMAT_LINES_FORMAT = _("tax_analysis.lines content is not a dict.")

    @classmethod
    def validate_tax_analysis_format(cls, tax: dict):
        if tax and not cls._has_required_keys(tax):
            raise ValidationError(cls.INVALID_TAX_ANALYSIS_JSON_FORMAT_NO_KEYS)

        # TODO when tax calculation rule is available then remove this condition
        if tax:
            lines = tax['lines']
            if not isinstance(lines, list):
                raise ValidationError(cls.INVALID_TAX_ANALYSIS_JSON_FORMAT_LINES_KEYS)

            for line in lines:
                if not isinstance(line, dict):
                    raise ValidationError(cls.INVALID_TAX_ANALYSIS_JSON_FORMAT_LINES_FORMAT)

    @classmethod
    def _has_required_keys(cls, tax: dict):
        keys = tax.keys()
        if 'total' not in keys or 'lines' not in keys:
            return False
        return True


class BaseInvoiceValidation(BaseModelValidation, UniqueCodeValidationMixin,
                            TaxAnalysisFormatValidationMixin, ObjectExistsValidationMixin, ABC):

    @classmethod
    def validate_create(cls, user, **data):
        cls.validate_unique_code_name(data.get('code', None))
        cls.validate_tax_analysis_format(data.get('tax_analysis', None))

    @classmethod
    def validate_update(cls, user, **data):
        cls.validate_object_exists(data.get('id', None))
        code = data.get('code', None)
        id_ = data.get('id', None)

        if code:
            cls.validate_unique_code_name(code, id_)

        cls.validate_tax_analysis_format(data.get('tax_analysis', None))



from .invoice_helpers import DEFAULT_TEST_INVOICE_PAYLOAD, create_test_invoice
from .invoice_line_item_helpers import DEFAULT_TEST_INVOICE_LINE_ITEM_PAYLOAD, create_test_invoice_line_item
from .invoice_payment_helpers import DEFAULT_TEST_INVOICE_PAYMENT_PAYLOAD, create_test_invoice_payment
from .bill_helpers import DEFAULT_TEST_BILL_PAYLOAD, create_test_bill
from .bill_line_item_helpers import DEFAULT_TEST_BILL_LINE_ITEM_PAYLOAD, create_test_bill_line_item
from .bill_payment_helpers import DEFAULT_TEST_BILL_PAYMENT_PAYLOAD, create_test_bill_payment
from .payment_invoice_helpers import (
    DEFAULT_TEST_PAYMENT_INVOICE_PAYLOAD,
    DEFAULT_TEST_DETAIL_PAYMENT_INVOICE_PAYLOAD,
    create_test_payment_invoice_without_details,
    create_test_payment_invoice_with_details
)



from invoice.validation.base import TaxAnalysisFormatValidationMixin, BaseInvoiceValidation
from invoice.validation.bill import BillModelValidation, BillItemStatus
from invoice.validation.billLineItem import BillLineItemModelValidation
from invoice.validation.invoice import InvoiceModelValidation, InvoiceItemStatus
from invoice.validation.invoiceLineItem import InvoiceLineItemModelValidation
from invoice.validation.invoicePayment import InvoicePaymentModelValidation



import graphene
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ValidationError

from core.gql.gql_mutations.base_mutation import BaseMutation, BaseCreateMutationMixin, BaseUpdateMutationMixin, \
    BaseHistoryModelCreateMutationMixin, BaseHistoryModelUpdateMutationMixin, BaseHistoryModelDeleteMutationMixin
from core.schema import OpenIMISMutation
from invoice.apps import InvoiceConfig
from invoice.gql.input_types import CreateBillPaymentInputType, UpdateBillPaymentInputType
from invoice.models import BillPayment, BillPaymentMutation


class CreateBillPaymentMutation(BaseHistoryModelCreateMutationMixin, BaseMutation):
    _mutation_class = "CreateBillPaymentMutation"
    _mutation_module = "invoice"
    _model = BillPayment

    @classmethod
    def _mutate(cls, user, **data):
        client_mutation_id = data.get("client_mutation_id")
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        p = cls.create_object(user=user, object_data=data)
        if p:
            BillPaymentMutation.object_mutated(
                user,
                client_mutation_id=client_mutation_id,
                bill_payment=p
            )

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_bill_payment_create_perms):
            raise ValidationError("mutation.authentication_required")

    class Input(CreateBillPaymentInputType):
        pass


class UpdateBillPaymentMutation(BaseHistoryModelUpdateMutationMixin, BaseMutation):
    _mutation_class = "UpdateBillPaymentMutation"
    _mutation_module = "invoice"
    _model = BillPayment

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_bill_payment_update_perms):
            raise ValidationError("mutation.authentication_required")

    class Input(UpdateBillPaymentInputType):
        pass


class DeleteBillPaymentMutation(BaseHistoryModelDeleteMutationMixin, BaseMutation):
    _mutation_class = "DeleteBillPaymentMutation"
    _mutation_module = "invoice"
    _model = BillPayment

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_bill_payment_delete_perms):
            raise ValidationError("mutation.authentication_required")

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.UUID)



import graphene
from django.contrib.auth.models import AnonymousUser
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError

from core.gql.gql_mutations.base_mutation import (
    BaseMutation,
    BaseHistoryModelCreateMutationMixin,
    BaseHistoryModelUpdateMutationMixin,
    BaseHistoryModelDeleteMutationMixin
)
from core.schema import OpenIMISMutation
from invoice.apps import InvoiceConfig
from invoice.gql.input_types import (
    CreatePaymentInvoiceInputType,
    CreatePaymentInvoiceWithDetailInputType,
    UpdatePaymentInvoiceInputType
)
from invoice.models import (
    PaymentInvoice,
    PaymentInvoiceMutation,
    DetailPaymentInvoice
)


class CreatePaymentInvoiceMutation(BaseHistoryModelCreateMutationMixin, BaseMutation):
    _mutation_class = "CreatePaymentInvoiceMutation"
    _mutation_module = "invoice"
    _model = PaymentInvoice

    @classmethod
    def _mutate(cls, user, **data):
        client_mutation_id = data.get("client_mutation_id")
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        payment_invoice = cls.create_object(user=user, object_data=data)
        if payment_invoice:
            PaymentInvoiceMutation.object_mutated(
                user,
                client_mutation_id=client_mutation_id,
                payment_invoice=payment_invoice
            )

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_invoice_payment_create_perms):
            raise ValidationError("mutation.authentication_required")

    class Input(CreatePaymentInvoiceInputType):
        pass


class CreatePaymentInvoiceWithDetailMutation(BaseHistoryModelCreateMutationMixin, BaseMutation):
    _mutation_class = "CreatePaymentInvoiceWithDetailMutation"
    _mutation_module = "invoice"
    _model = PaymentInvoice

    @classmethod
    def _mutate(cls, user, **data):
        client_mutation_id = data.get("client_mutation_id")
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')
        status, subject_id, subject_type = cls._get_field_for_detail(data)
        payment_invoice = cls.create_object(user=user, object_data=data)
        if payment_invoice:
            PaymentInvoiceMutation.object_mutated(
                user,
                client_mutation_id=client_mutation_id,
                payment_invoice=payment_invoice
            )
            cls._create_payment_detail(user, data, payment_invoice, status, subject_id, subject_type)

    @classmethod
    def _get_field_for_detail(cls, data):
        status = data.pop('status')
        subject_id = data.pop('subject_id')
        subject_type = data.pop('subject_type')
        return status, subject_id, subject_type

    @classmethod
    def _create_payment_detail(cls, user, data, payment, status, subject_id, subject_type):
        payment_detail = cls._build_payment_detail(data, payment, status, subject_id, subject_type)
        detail_payment_invoice = DetailPaymentInvoice(**payment_detail)
        detail_payment_invoice.save(username=user.username)

    @classmethod
    def _build_payment_detail(cls, data, payment, status, subject_id, subject_type):
        return {
            "status": status,
            "payment_id": payment.id,
            "subject_type": cls._convert_content_type(subject_type),
            "subject_id": subject_id,
            "fees": data['fees'],
            "amount": data['amount_received'],
        }

    @classmethod
    def _convert_content_type(cls, subject_type):
        return ContentType.objects.get(model__iexact=subject_type)

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_invoice_payment_create_perms):
            raise ValidationError("mutation.authentication_required")

    class Input(CreatePaymentInvoiceWithDetailInputType):
        pass


class UpdatePaymentInvoiceMutation(BaseHistoryModelUpdateMutationMixin, BaseMutation):
    _mutation_class = "UpdatePaymentInvoiceMutation"
    _mutation_module = "invoice"
    _model = PaymentInvoice

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_invoice_payment_update_perms):
            raise ValidationError("mutation.authentication_required")

    class Input(UpdatePaymentInvoiceInputType):
        pass


class DeletePaymentInvoiceMutation(BaseHistoryModelDeleteMutationMixin, BaseMutation):
    _mutation_class = "DeletePaymentInvoiceMutation"
    _mutation_module = "invoice"
    _model = PaymentInvoice

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_invoice_payment_delete_perms):
            raise ValidationError("mutation.authentication_required")

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.UUID)



import decimal
from typing import Union, List

from invoice.models import Bill, BillItem
from core.services import BaseService
from invoice.services.billLineItem import BillLineItemService
from core.services.utils import get_generic_type
from invoice.validation.bill import BillModelValidation, BillItemStatus
from core.signals import *


class BillService(BaseService):
    OBJECT_TYPE = Bill

    def __init__(self, user, validation_class: BillModelValidation = BillModelValidation):
        super().__init__(user, validation_class)
        self.validation_class = validation_class

    def _base_payload_adjust(self, bill_data):
        return self._evaluate_generic_types(bill_data)

    def bill_validate_items(self, bill: Bill):
        # TODO: Implement after calculation rules available
        pass

    def bill_match_items(self, bill: Bill) \
            -> Dict[str, Union[BillItemStatus, Dict[BillItem, List[BillItemStatus]]]]:
        """
        Check if items related to bill are valid.
        @param bill: Bill object
        @return: Dict with two keys, 'subject_status' containing information if bill subject is valid and
        'line_items', containing information about statuses of lines connected to bill items.
        """
        match_result = {
            'subject': self.validation_class.validate_subject(bill),
            'line_items': self.validation_class.validate_line_items(bill)
        }
        return match_result

    def billTaxCalculation(self, bill: Bill):
        # TODO: Implement after calculation rules available
        pass

    @classmethod
    @register_service_signal('signal_after_invoice_module_bill_create_service')
    def bill_create(cls, **kwargs):
        convert_results = kwargs.get('convert_results', {})
        if 'bill_data' in convert_results and 'bill_data_line' in convert_results:
            user = convert_results['user']
            # save in database this invoice and invoice line item
            bill_line_items = convert_results['bill_data_line']
            bill_service = BillService(user=user)
            bill_line_item_service = BillLineItemService(user=user)
            result_bill = bill_service.create(convert_results['bill_data'])
            if result_bill["success"] is True:
                bill_update = {
                    "id": result_bill["data"]["id"],
                    "amount_net": decimal.Decimal(0),
                    "amount_total": decimal.Decimal(0),
                    "amount_discount": decimal.Decimal(0),
                }
                for bill_line_item in bill_line_items:
                    bill_line_item["bill_id"] = result_bill["data"]["id"]
                    result_bill_line = bill_line_item_service.create(bill_line_item)
                    if result_bill_line["success"] is True:
                        bill_update["amount_net"] += decimal.Decimal(result_bill_line["data"]["amount_net"])
                        bill_update["amount_total"] += decimal.Decimal(result_bill_line["data"]["amount_total"])
                        bill_update["amount_discount"] += decimal.Decimal(0) \
                            if not result_bill_line["data"]["discount"] \
                            else decimal.Decimal(result_bill_line["data"]["discount"])
                generated_bill = bill_service.update(bill_update)
                return generated_bill

    def _evaluate_generic_types(self, bill_data):
        if 'subject_type' in bill_data.keys():
            bill_data['subject_type'] = get_generic_type(bill_data['subject_type'])

        if 'thirdparty_type' in bill_data.keys():
            bill_data['thirdparty_type'] = get_generic_type(bill_data['thirdparty_type'])

        return bill_data



import graphene

from core.schema import OpenIMISMutation


class CreatePaymentInputType(OpenIMISMutation.Input):
    status = graphene.Int(required=False)

    code_ext = graphene.String(required=False, max_length=32)
    code_tp = graphene.String(required=False, max_length=32)
    code_receipt = graphene.String(required=False, max_length=32)

    payment_origin = graphene.String(required=False, max_length=32)

    label = graphene.String(required=False, max_length=32)

    invoice_id = graphene.UUID(required=False)

    amount_payed = graphene.Decimal(max_digits=18, decimal_places=2, required=False)
    fees = graphene.Decimal(max_digits=18, decimal_places=2, required=False)
    amount_received = graphene.Decimal(max_digits=18, decimal_places=2, required=False)

    date_payment = graphene.Date(required=False)

    date_valid_from = graphene.Date(required=False)
    date_valid_to = graphene.Date(required=False)
    json_ext = graphene.types.json.JSONString(required=False)


class UpdatePaymentInputType(CreatePaymentInputType):
    id = graphene.UUID(required=True)


class CreateInvoiceEventType(OpenIMISMutation.Input):
    invoice_id = graphene.UUID(required=True)
    event_type = graphene.Int(required=True)
    message = graphene.String(required=True)


class UpdateInvoiceEventType(CreateInvoiceEventType):
    id = graphene.UUID(required=True)


class CreateBillPaymentInputType(OpenIMISMutation.Input):
    status = graphene.Int(required=False)

    code_ext = graphene.String(required=False, max_length=32)
    code_tp = graphene.String(required=False, max_length=32)
    code_receipt = graphene.String(required=False, max_length=32)

    payment_origin = graphene.String(required=False, max_length=32)

    label = graphene.String(required=False, max_length=32)

    bill_id = graphene.UUID(required=False)

    amount_payed = graphene.Decimal(max_digits=18, decimal_places=2, required=False)
    fees = graphene.Decimal(max_digits=18, decimal_places=2, required=False)
    amount_received = graphene.Decimal(max_digits=18, decimal_places=2, required=False)

    date_payment = graphene.Date(required=False)

    date_valid_from = graphene.Date(required=False)
    date_valid_to = graphene.Date(required=False)
    json_ext = graphene.types.json.JSONString(required=False)


class UpdateBillPaymentInputType(CreateBillPaymentInputType):
    id = graphene.UUID(required=True)


class CreateBillEventType(OpenIMISMutation.Input):
    bill_id = graphene.UUID(required=True)
    event_type = graphene.Int(required=True)
    message = graphene.String(required=True)


class UpdateBillEventType(CreateBillEventType):
    id = graphene.UUID(required=True)


class CreatePaymentInvoiceInputType(OpenIMISMutation.Input):
    reconciliation_status = graphene.Int(required=False)

    code_ext = graphene.String(required=False, max_length=32)
    code_tp = graphene.String(required=False, max_length=32)
    code_receipt = graphene.String(required=False, max_length=32)

    label = graphene.String(required=False, max_length=32)

    fees = graphene.Decimal(max_digits=18, decimal_places=2, required=False)
    amount_received = graphene.Decimal(max_digits=18, decimal_places=2, required=False)

    date_payment = graphene.Date(required=False)

    payment_origin = graphene.String(required=False, max_length=32)

    payer_ref = graphene.String(required=True, max_length=32)
    payer_name = graphene.String(required=False, max_length=32)

    date_valid_from = graphene.Date(required=False)
    date_valid_to = graphene.Date(required=False)
    json_ext = graphene.types.json.JSONString(required=False)


class CreatePaymentInvoiceWithDetailInputType(CreatePaymentInvoiceInputType):
    status = graphene.Int(required=True)
    subject_id = graphene.UUID(required=True)
    subject_type = graphene.String(required=True, max_length=32)


class UpdatePaymentInvoiceInputType(CreatePaymentInvoiceInputType):
    id = graphene.UUID(required=True)


class CreateDetailPaymentInvoiceInputType(OpenIMISMutation.Input):
    status = graphene.Int(required=False)

    payment_id = graphene.UUID(required=False)

    subject_id = graphene.UUID(required=False)
    subject_type = graphene.String(required=False, max_length=32)

    fees = graphene.Decimal(max_digits=18, decimal_places=2, required=False)
    amount = graphene.Decimal(max_digits=18, decimal_places=2, required=False)

    date_valid_from = graphene.Date(required=False)
    date_valid_to = graphene.Date(required=False)
    json_ext = graphene.types.json.JSONString(required=False)


class UpdateDetailPaymentInvoiceInputType(CreateDetailPaymentInvoiceInputType):
    id = graphene.UUID(required=True)



# Generated by Django 3.0.14 on 2022-04-11 13:58

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('invoice', '0007_auto_20220411_1053'),
    ]

    operations = [
        migrations.AlterField(
            model_name='detailpaymentinvoice',
            name='payment',
            field=models.ForeignKey(db_column='PaymentUUID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='invoice_payments', to='invoice.PaymentInvoice'),
        ),
    ]



from django.db import transaction
from policy.test_helpers import create_test_policy
from product.test_helpers import create_test_product

from core.forms import User
from core.test_helpers import compare_dicts
from django.test import TestCase

from invoice.models import InvoiceLineItem
from contract.tests.helpers import create_test_contract
from policyholder.tests.helpers import create_test_policy_holder
from insuree.test_helpers import create_test_insuree

from invoice.services.invoiceLineItem import InvoiceLineItemService
from invoice.tests.helpers import create_test_invoice


class ServiceTestInvoiceLineItems(TestCase):
    BASE_TEST_INVOICE_LINE_ITEM_PAYLOAD = {
        'code': 'LineItem1',
        # 'line_type': None,
        # 'line_id': None,
        # 'invoiceId': None,
        'description': 'description_str',
        'details': '{"test_int": 1, "test_txt": "some_str"}',
        'ledger_account': 'account',
        'quantity': 10,
        'unit_price': 10.5,
        'discount': 15.5,
        'tax_rate': None,
        'tax_analysis': {'lines': [{'code': 'c', 'label': 'l', 'base': '0.1', 'amount': '2.00'}], 'total': '2.0'},
    }

    BASE_TEST_UPDATE_LINE_ITEM_PAYLOAD = {
        'code': 'LineItem2',
        'description': 'description_str2',
        'details': '{"test_int": 1, "test_txt": "some_str"}',
        'ledger_account': 'account2',
        'quantity': 12,
        'unit_price': 10,
        'discount': 20,
        'tax_rate': None,
        'tax_analysis': {'lines': [{'code': 'c', 'label': 'l', 'base': '0.1', 'amount': '1.00'}], 'total': '1.0'},
    }

    BASE_EXPECTED_CREATE_RESPONSE = {
        "success": True,
        "message": "Ok",
        "detail": "",
        "data": {
            'code': 'LineItem1',
            'line_type': None,
            'line_id': None,
            'invoice': None,
            'amount_total': 91.5,
            'amount_net': 89.5,
            'description': 'description_str',
            'details': {"test_int": 1, "test_txt": "some_str"},
            'ledger_account': 'account',
            'quantity': 10,
            'unit_price': 10.5,
            'discount': 15.5,
            'tax_rate': None,
            'tax_analysis': {'lines': [{'code': 'c', 'label': 'l', 'base': '0.1', 'amount': '2.00'}], 'total': '2.0'}
        },
    }

    BASE_EXPECTED_UPDATE_RESPONSE = {
        "success": True,
        "message": "Ok",
        "detail": "",
        "data": {
            'code': 'LineItem2',
            'amount_total': 101.0,
            'amount_net': 100.0,
            'description': 'description_str2',
            'details': {"test_int": 1, "test_txt": "some_str"},
            'ledger_account': 'account2',
            'quantity': 12,
            'unit_price': 10.0,
            'discount': 20.0,
            'tax_analysis': {'lines': [{'code': 'c', 'label': 'l', 'base': '0.1', 'amount': '1.00'}], 'total': '1.0'},
            'line_id': None,
            'line_type': None
        },
    }

    @classmethod
    def setUpClass(cls):
        super(ServiceTestInvoiceLineItems, cls).setUpClass()
        if not User.objects.filter(username='admin_invoice').exists():
            User.objects.create_superuser(username='admin_invoice', password='S\/pe®Pąßw0rd™')

        cls.policy_holder = create_test_policy_holder()
        cls.contract = create_test_contract(cls.policy_holder)
        cls.user = User.objects.filter(username='admin').first()
        cls.insuree = create_test_insuree(with_family=True)
        cls.line_item_service = InvoiceLineItemService(cls.user)
        cls.invoice = create_test_invoice(cls.contract, cls.insuree)
        cls.product = create_test_product("TestC0d3", custom_props={"insurance_period": 12})
        cls.policy = create_test_policy(
            product=cls.product,
            insuree=cls.insuree
        )

        cls.BASE_TEST_INVOICE_LINE_ITEM_PAYLOAD['line'] = cls.policy
        cls.BASE_TEST_INVOICE_LINE_ITEM_PAYLOAD['invoice'] = cls.invoice

    def test_line_items_create(self):
        with transaction.atomic():
            payload = self.BASE_TEST_INVOICE_LINE_ITEM_PAYLOAD.copy()
            payload['invoice'] = self.invoice

            expected_response = self.BASE_EXPECTED_CREATE_RESPONSE.copy()
            expected_response['data']['invoice'] = str(self.invoice.pk)

            response = self.line_item_service.create(payload)

            truncated_output = response
            truncated_output['data'] = {k: v for k, v in truncated_output['data'].items()
                                        if k in expected_response['data'].keys()}

            line_item = InvoiceLineItem.objects.filter(code=payload['code']).first()
            expected_response['data']['line_type'] = line_item.line_type.id
            expected_response['data']['line_id'] = self.policy.pk

            self.assertTrue(compare_dicts(expected_response, response))
            InvoiceLineItem.objects.filter(code=payload['code']).delete()

    def test_line_items_update(self):
        self.maxDiff = None
        with transaction.atomic():
            create_payload = self.BASE_TEST_INVOICE_LINE_ITEM_PAYLOAD.copy()
            create_payload['invoice'] = self.invoice

            expected_response = self.BASE_EXPECTED_UPDATE_RESPONSE.copy()
            expected_response['data']['invoice'] = str(self.invoice.pk)

            self.line_item_service.create(create_payload)

            update_payload = self.BASE_TEST_UPDATE_LINE_ITEM_PAYLOAD.copy()
            update_payload['id'] = InvoiceLineItem.objects.filter(code=create_payload['code']).first().id

            response = self.line_item_service.update(update_payload)
            truncated_output = response
            truncated_output['data'] = {k: v for k, v in truncated_output['data'].items()
                                        if k in expected_response['data'].keys()}

            line_item = InvoiceLineItem.objects.filter(code=update_payload['code']).first()

            expected_response['data']['line_type'] = line_item.line_type.id
            expected_response['data']['line_id'] = str(self.policy.pk)

            self.assertTrue(compare_dicts(expected_response, response))
            InvoiceLineItem.objects.filter(code=update_payload['code']).delete()

    def test_line_items_delete(self):
        with transaction.atomic():
            payload = self.BASE_TEST_INVOICE_LINE_ITEM_PAYLOAD.copy()
            expected_response = {
                "success": True,
                "message": "Ok",
                "detail": "",
            }

            response = self.line_item_service.create(payload)
            response = self.line_item_service.delete(response['data'])
            self.assertDictEqual(expected_response, response)
            InvoiceLineItem.objects.filter(code=payload['code']).delete()



from core.signals import bind_service_signal
from core.service_signals import ServiceSignalBindType
from django.contrib.contenttypes.models import ContentType
from invoice.models import InvoiceLineItem
from invoice.services import InvoiceService


def bind_service_signals():
    bind_service_signal(
        'convert_to_invoice',
        check_invoice_exist,
        bind_type=ServiceSignalBindType.BEFORE
    )
    bind_service_signal(
        'convert_to_invoice',
        save_invoice_in_db,
        bind_type=ServiceSignalBindType.AFTER
    )


def check_invoice_exist(**kwargs):
    function_arguments = kwargs.get('data')[1]
    instance = function_arguments.get('instance', None)
    content_type_policy = ContentType.objects.get_for_model(instance.__class__)
    invoices = InvoiceLineItem.objects.filter(line_type=content_type_policy, line_id=instance.id)
    if invoices.count() == 0:
        return True


def save_invoice_in_db(**kwargs):
    InvoiceService.invoice_create(**kwargs)



import re

from django.contrib.contenttypes.models import ContentType
from django.db.models import Q

from invoice.models import (
    DetailPaymentInvoice,
    Invoice,
    InvoiceLineItem,
    Bill,
    BillItem
)

camel_pat = re.compile(r'([A-Z])')
under_pat = re.compile(r'_([a-z])')


def camel_to_underscore(name):
    return camel_pat.sub(lambda x: '_' + x.group(1).lower(), name)


def underscore_to_camel(name):
    return under_pat.sub(lambda x: x.group(1).upper(), name)


def resolve_payment_details(payment_invoice):
    invoice_list = _retrieve_list_of_invoices_bills(payment_invoice, Invoice)
    invoice_line_item_list = _retrieve_list_of_invoices_bills(payment_invoice, InvoiceLineItem)
    invoices = Invoice.objects.filter(Q(id__in=invoice_list) | Q(line_items__id__in=invoice_line_item_list))
    bill_list = _retrieve_list_of_invoices_bills(payment_invoice, Bill)
    bill_line_item_list = _retrieve_list_of_invoices_bills(payment_invoice, BillItem)
    bills = Bill.objects.filter(Q(id__in=bill_list) | Q(line_items_bill__id__in=bill_line_item_list))
    return invoices, bills


def _retrieve_list_of_invoices_bills(payment_invoice, model_type):
    return list(DetailPaymentInvoice.objects.filter(
        payment=payment_invoice,
        subject_type=ContentType.objects.get_for_model(model_type).id
    ).values_list('subject_id', flat=True))



import graphene
from django.contrib.auth.models import AnonymousUser
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from pandas import DataFrame

from core.gql.export_mixin import ExportableQueryMixin
from core.schema import OrderedDjangoFilterConnectionField
from core.utils import append_validity_filter
from invoice.apps import InvoiceConfig
from invoice.gql.gql_types.bill_types import BillGQLType
from invoice.models import Bill
import graphene_django_optimizer as gql_optimizer
from django.apps import apps
try:
    Policy = apps.get_model('policy','Policy')
except:
    Policy = {}


def patch_subjects(bills_df: DataFrame):
    subject_df = bills_df[['subject_type', 'subject_id']].drop_duplicates()
    bills_df['subject_class'] = 'undefined'
    for subject in subject_df['subject_type'].unique():
        model = ContentType.objects.get(id=subject).model_class()
        entities_for_model = subject_df[subject_df['subject_type'] == subject]['subject_id']
        if model == Policy:
            subject_names = model.objects \
                .filter(id__in=entities_for_model) \
                .values('id', 'family__head_insuree__chf_id',
                        'family__head_insuree__last_name', 'family__head_insuree__other_names')
            names = {
                str(x['id']): F"{x['family__head_insuree__other_names']} {x['family__head_insuree__last_name']} ({x['family__head_insuree__chf_id']})"
                for x in subject_names
            }
            updated = bills_df['subject_type'] == subject
            bills_df.loc[updated, 'subject_id'] = bills_df[updated]['subject_id'].apply(lambda x: names[x])
        bills_df.loc[bills_df['subject_type'] == subject, 'subject_class'] = model.__name__
        bills_df['subject_type'] = bills_df.pop('subject_class')
    return bills_df


class BillQueryMixin(ExportableQueryMixin):
    export_patches = {
        'bill': [
            patch_subjects
        ]
    }

    exportable_fields = ['bill']
    bill = OrderedDjangoFilterConnectionField(
        BillGQLType,
        orderBy=graphene.List(of_type=graphene.String),
        dateValidFrom__Gte=graphene.DateTime(),
        dateValidTo__Lte=graphene.DateTime(),
        applyDefaultValidityFilter=graphene.Boolean(),
        client_mutation_id=graphene.String(),
        subject_type_filter=graphene.String(),
        thirdparty_type_filter=graphene.String(),
    )

    def resolve_bill(self, info, **kwargs):
        BillQueryMixin._check_permissions(info.context.user)
        filters = []
        filters += append_validity_filter(**kwargs)

        client_mutation_id = kwargs.get("client_mutation_id", None)
        if client_mutation_id:
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        subject_type = kwargs.pop("subject_type_filter", None)
        if subject_type:
            filters.append(Q(subject_type__model=subject_type))

        thirdparty_type = kwargs.get("thirdparty_type_filter", None)
        if thirdparty_type:
            filters.append(Q(thirdparty_type__model=thirdparty_type))

        qs = Bill.objects.filter(*filters)
        if InvoiceConfig.bill_user_filter:
            qs = InvoiceConfig.bill_user_filter(qs, info.context.user)

        return gql_optimizer.query(qs, info)

    @staticmethod
    def _check_permissions(user):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_bill_search_perms):
            raise PermissionError("Unauthorized")



from .invoice_schema import *
from .invoice_line_item_schema import *
from .invoice_payment_schema import *
from .invoice_event_schema import *
from .payment_invoice_schema import *
from .detail_payment_invoice_schema import *



import logging

from django.db import migrations

from core.utils import insert_role_right_for_system

logger = logging.getLogger(__name__)


def add_rights(apps, schema_editor):
    insert_role_right_for_system(64, 155101, apps)  # Invoice search
    insert_role_right_for_system(64, 155102, apps)  # Invoice create
    insert_role_right_for_system(64, 155103, apps)  # Invoice update
    insert_role_right_for_system(64, 155104, apps)  # Invoice delete
    insert_role_right_for_system(64, 155109, apps)  # Invoice amend
    insert_role_right_for_system(64, 155201, apps)  # Invoice payment search
    insert_role_right_for_system(64, 155202, apps)  # Invoice payment create
    insert_role_right_for_system(64, 155203, apps)  # Invoice payment update
    insert_role_right_for_system(64, 155204, apps)  # Invoice payment delete
    insert_role_right_for_system(64, 155206, apps)  # Invoice payment refund
    insert_role_right_for_system(64, 155301, apps)  # Invoice Event search
    insert_role_right_for_system(64, 155302, apps)  # Invoice Event create
    insert_role_right_for_system(64, 155303, apps)  # Invoice Event update
    insert_role_right_for_system(64, 155304, apps)  # Invoice Event delete
    insert_role_right_for_system(64, 155306, apps)  # Invoice Event message
    insert_role_right_for_system(64, 155307, apps)  # Invoice Event delete my message
    insert_role_right_for_system(64, 155308, apps)  # Invoice Event delete all messages
    insert_role_right_for_system(64, 156101, apps)  # Bill search
    insert_role_right_for_system(64, 156102, apps)  # Bill create
    insert_role_right_for_system(64, 156103, apps)  # Bill update
    insert_role_right_for_system(64, 156104, apps)  # Bill delete
    insert_role_right_for_system(64, 156109, apps)  # Bill amend
    insert_role_right_for_system(64, 156201, apps)  # Bill Payment search
    insert_role_right_for_system(64, 156202, apps)  # Bill Payment create
    insert_role_right_for_system(64, 156203, apps)  # Bill Payment update
    insert_role_right_for_system(64, 156204, apps)  # Bill Payment delete
    insert_role_right_for_system(64, 156206, apps)  # Bill Payment refund
    insert_role_right_for_system(64, 156301, apps)  # Bill Event search
    insert_role_right_for_system(64, 156302, apps)  # Bill Event create
    insert_role_right_for_system(64, 156303, apps)  # Bill Event update
    insert_role_right_for_system(64, 156304, apps)  # Bill Event delete
    insert_role_right_for_system(64, 156306, apps)  # Bill Event create message
    insert_role_right_for_system(64, 156307, apps)  # Bill Event delete my message
    insert_role_right_for_system(64, 156308, apps)  # Bill Event delete all messages


class Migration(migrations.Migration):
    dependencies = [
        ('invoice', '0003_auto_20211203_1053')
    ]

    operations = [
        migrations.RunPython(add_rights),
    ]



from json import JSONDecodeError

import json

from invoice.models import InvoiceLineItem
from core.services import BaseService
from core.services.utils import get_generic_type
from invoice.validation.invoiceLineItem import InvoiceLineItemModelValidation

import logging

logger = logging.getLogger(__name__)


class InvoiceLineItemService(BaseService):

    OBJECT_TYPE = InvoiceLineItem

    def __init__(self, user, validation_class: InvoiceLineItemModelValidation = InvoiceLineItemModelValidation):
        super().__init__(user, validation_class)

    def _base_payload_adjust(self, invoice_data):
        adjusted_generics = self._evaluate_generic_types(invoice_data)
        adjusted_details = self._adjust_details_field(adjusted_generics)
        adjusted_calculations = self._calculate_payload_values(adjusted_details)
        return adjusted_calculations

    def _evaluate_generic_types(self, invoice_data):
        if 'line_type' in invoice_data.keys():
            invoice_data['line_type'] = get_generic_type(invoice_data['line_type'])
        return invoice_data

    def _calculate_payload_values(self, invoice_data):
        if 'amount_net' not in invoice_data.keys():
            invoice_data['amount_net'] = self.__calculate_net(invoice_data)
        if 'amount_total' not in invoice_data.keys():
            invoice_data['amount_total'] = self.__calculate_total(invoice_data)
        return invoice_data

    def _adjust_details_field(self, invoice_data):
        if invoice_data.get("details", None):
            details = invoice_data.get('details')
            if isinstance(details, dict):
                return invoice_data
            elif isinstance(details, str):
                try:
                    data = json.loads(details)
                except JSONDecodeError as e:
                    logger.exception("Failed to parse invoice line item "
                                     f"details {details}, conent will be saved in text"
                                     f"field")
                    data = {'text': details}
                invoice_data['details'] = data
            else:
                raise TypeError(f"Invalid type for InvoiceLineItem.details "
                                f"{type(details)}. Expected str or dict.")
        return invoice_data

    def __calculate_net(self, invoice_data):
        quantity = invoice_data.get('quantity', 0)
        unit_price = invoice_data.get('unit_price', 0)
        discount = invoice_data.get('discount', 0)
        return (quantity * unit_price) - discount

    def __calculate_total(self, invoice_data):
        tax_total = self.__get_tax_total(invoice_data)
        return invoice_data.get('amount_net', 0) + tax_total

    def __get_tax_total(self, invoice_data):
        tax_data = invoice_data.get('tax_analysis', {'total': 0})
        if isinstance(tax_data, str):
            tax_data = json.loads(tax_data)
        if 'total' not in tax_data.keys():
            logger.warning("Invalid 'tax_analysis' format, total field not found")
            return 0
        else:
            return float(tax_data['total'])



import logging

import graphene
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ValidationError
from graphql import GraphQLError

from core.gql.gql_mutations.base_mutation import BaseMutation, BaseHistoryModelDeleteMutationMixin
from core.schema import OpenIMISMutation
from invoice.apps import InvoiceConfig
from invoice.models import Invoice
from invoice.services import InvoiceService

logger = logging.getLogger(__name__)


class DeleteInvoiceMutation(BaseHistoryModelDeleteMutationMixin, BaseMutation):
    _mutation_class = "DeleteInvoiceMutation"
    _mutation_module = "invoice"
    _model = Invoice

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_invoice_delete_perms):
            raise ValidationError("mutation.authentication_required")

    class Input(OpenIMISMutation.Input):
        uuids = graphene.List(graphene.UUID)


class GenerateTimeframeInvoices(BaseMutation):
    _model = Invoice
    _invoice_service_class: InvoiceService = InvoiceService
    _mutation_module = 'invoice'
    _mutation_class = 'GenerateTimeframeInvoices'

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                InvoiceConfig.gql_invoice_create_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')

        from_date, to_date = data.get("date_from", None), data.get("date_to", None)

        if not from_date or not to_date:
            raise GraphQLError('To generate invoices for a time period, '
                               'it is necessary to specify from_date and to_date')

        output = cls._generate_timeframe_invoices(user, from_date, to_date)

        return [] if output is None else f"Error during invoice generation: {output}"

    @classmethod
    def _generate_timeframe_invoices(cls, user, from_date, to_date):
        try:
            # invoice_creation_from_calculation doesn't have implementation,
            # it sends a signal, if a binded function throws an exception it's message is returned.
            service = cls._invoice_service_class(user)
            service.invoice_creation_from_calculation(user=user, from_date=from_date, to_date=to_date)
        except Exception as e:
            logger.exception(F"Exception occurred during invoice generation. Details: {e}")
            return str(e)
        return None

    class Input(OpenIMISMutation.Input):
        date_from = graphene.Date(required=False)
        date_to = graphene.Date(required=False)




from typing import Union, List

from invoice.models import Invoice, InvoiceLineItem
from core.services import BaseService
from core.services.utils import get_generic_type
from invoice.services.invoiceLineItem import InvoiceLineItemService
from invoice.validation.invoice import InvoiceModelValidation, InvoiceItemStatus
from core.signals import *


class InvoiceService(BaseService):
    OBJECT_TYPE = Invoice

    def __init__(self, user, validation_class: InvoiceModelValidation = InvoiceModelValidation):
        super().__init__(user, validation_class)
        self.validation_class = validation_class

    def _base_payload_adjust(self, invoice_data):
        return self._evaluate_generic_types(invoice_data)

    def invoice_validate_items(self, invoice: Invoice):
        # TODO: Implement after calculation rules available
        pass

    def invoice_match_items(self, invoice: Invoice) \
            -> Dict[str, Union[InvoiceItemStatus, Dict[InvoiceLineItem, List[InvoiceItemStatus]]]]:
        """
        Check if items related to invoice are valid.
        @param invoice: Invoice object
        @return: Dict with two keys, 'subject_status' containing information if invoice subject is valid and
        'line_items', containing information about statuses of lines connected to invoice items.
        """
        match_result = {
            'subject': self.validation_class.validate_subject(invoice),
            'line_items': self.validation_class.validate_line_items(invoice)
        }
        return match_result

    def invoiceTaxCalculation(self, invoice: Invoice):
        # TODO: Implement after calculation rules available
        pass

    @classmethod
    @register_service_signal('invoice_creation_from_calculation')
    def invoice_creation_from_calculation(cls, user, from_date, to_date):
        """
        It sends the invoice_creation_from_calculation signal which should inform the
        relevant calculation rule that invoices need to be generated.
        """
        pass

    def _evaluate_generic_types(self, invoice_data):
        if 'subject_type' in invoice_data.keys():
            invoice_data['subject_type'] = get_generic_type(invoice_data['subject_type'])

        if 'thirdparty_type' in invoice_data.keys():
            invoice_data['thirdparty_type'] = get_generic_type(invoice_data['thirdparty_type'])

        return invoice_data

    @classmethod
    @register_service_signal('signal_after_invoice_module_invoice_create_service')
    def invoice_create(cls, **kwargs):
        convert_results = kwargs.get('result', {})
        if 'invoice_data' in convert_results and 'invoice_data_line' in convert_results:
            user = convert_results['user']
            # save in database this invoice and invoice line item
            invoice_line_items = convert_results['invoice_data_line']
            invoice_service = InvoiceService(user=user)
            invoice_line_item_service = InvoiceLineItemService(user=user)
            result_invoice = invoice_service.create(convert_results['invoice_data'])
            if result_invoice["success"] is True:
                invoice_update = {
                    "id": result_invoice["data"]["id"],
                    "amount_net": 0,
                    "amount_total": 0,
                    "amount_discount": 0
                }
                for invoice_line_item in invoice_line_items:
                    invoice_line_item["invoice_id"] = result_invoice["data"]["id"]
                    result_invoice_line = invoice_line_item_service.create(invoice_line_item)
                    if result_invoice_line["success"] is True:
                        invoice_update["amount_net"] += float(result_invoice_line["data"]["amount_net"])
                        invoice_update["amount_total"] += float(result_invoice_line["data"]["amount_total"])
                        invoice_update["amount_discount"] += 0 if result_invoice_line["data"]["discount"] else \
                            result_invoice_line["data"]["discount"]
                generated_invoice = invoice_service.update(invoice_update)
                return generated_invoice
